_handle_types = 'getField getStatic putField putStatic invokeVirtual invokeStatic invokeSpecial newInvokeSpecial invokeInterface'.split()
handle_codes = dict(zip(_handle_types, range(1,10)))
handle_rcodes = {v:k for k,v in handle_codes.items()}

newarr_rcodes = ([None]*4) + 'boolean char float double byte short int long'.split()
newarr_codes = dict(zip('boolean char float double byte short int long'.split(), range(4,12)))

vt_rcodes = ['Top','Integer','Float','Double','Long','Null','UninitializedThis','Object','Uninitialized']
vt_codes = {k:i for i,k in enumerate(vt_rcodes)}

et_rtags = dict(zip(map(ord, 'BCDFIJSZsec@['), 'byte char double float int long short boolean string enum class annotation array'.split()))
et_tags = {v:k for k,v in et_rtags.items()}



import itertools

from .. import excepttypes, ssa_types
from ..constraints import IntConstraint, ObjectConstraint, returnOrThrow, return_, throw

from . import bitwise_util
from .base import BaseOp

def getNewRange(w, zmin, zmax):
    HN = 1 << w-1
    zmin = zmin + HN
    zmax = zmax + HN
    split = (zmin>>w != zmax>>w)

    if split:
        return return_(IntConstraint.range(w, -HN, HN-1))
    else:
        N = 1<<w
        return return_(IntConstraint.range(w, (zmin % N)-HN, (zmax % N)-HN))

class IAdd(BaseOp):
    def __init__(self, parent, args):
        super(IAdd, self).__init__(parent, args)
        self.rval = parent.makeVariable(args[0].type, origin=self)

    def propagateConstraints(self, x, y):
        return getNewRange(x.width, x.min+y.min, x.max+y.max)

class IMul(BaseOp):
    def __init__(self, parent, args):
        super(IMul, self).__init__(parent, args)
        self.rval = parent.makeVariable(args[0].type, origin=self)

    def propagateConstraints(self, x, y):
        vals = x.min*y.min, x.min*y.max, x.max*y.min, x.max*y.max
        return getNewRange(x.width, min(vals), max(vals))

class ISub(BaseOp):
    def __init__(self, parent, args):
        super(ISub, self).__init__(parent, args)
        self.rval = parent.makeVariable(args[0].type, origin=self)

    def propagateConstraints(self, x, y):
        return getNewRange(x.width, x.min-y.max, x.max-y.min)

#############################################################################################
class IAnd(BaseOp):
    def __init__(self, parent, args):
        BaseOp.__init__(self, parent, args)
        self.rval = parent.makeVariable(args[0].type, origin=self)

    def propagateConstraints(self, x, y):
        return return_(bitwise_util.propagateAnd(x,y))

class IOr(BaseOp):
    def __init__(self, parent, args):
        BaseOp.__init__(self, parent, args)
        self.rval = parent.makeVariable(args[0].type, origin=self)

    def propagateConstraints(self, x, y):
        return return_(bitwise_util.propagateOr(x,y))

class IXor(BaseOp):
    def __init__(self, parent, args):
        BaseOp.__init__(self, parent, args)
        self.rval = parent.makeVariable(args[0].type, origin=self)

    def propagateConstraints(self, x, y):
        return return_(bitwise_util.propagateXor(x,y))

#############################################################################################
# Shifts currently only propogate ranges in the case where the shift is a known constant
# TODO - make this handle the general case
def getMaskedRange(x, bits):
    assert bits < x.width
    y = IntConstraint.const(x.width, (1<<bits) - 1)
    x = bitwise_util.propagateAnd(x,y)

    H = 1<<(bits-1)
    M = 1<<bits

    m1 = x.min if (x.max <= H-1) else -H
    m2 = x.max if (x.min >= -H) else H-1
    return m1, m2

class IShl(BaseOp):
    def __init__(self, parent, args):
        BaseOp.__init__(self, parent, args)
        self.rval = parent.makeVariable(args[0].type, origin=self)

    def propagateConstraints(self, x, y):
        if y.min < y.max:
            return return_(IntConstraint.bot(x.width))
        shift = y.min % x.width
        if not shift:
            return return_(x)
        m1, m2 = getMaskedRange(x, x.width - shift)
        return return_(IntConstraint.range(x.width, m1<<shift, m2<<shift))

class IShr(BaseOp):
    def __init__(self, parent, args):
        BaseOp.__init__(self, parent, args)
        self.rval = parent.makeVariable(args[0].type, origin=self)

    def propagateConstraints(self, x, y):
        if y.min < y.max:
            return return_(IntConstraint.range(x.width, min(x.min, 0), max(x.max, 0)))
        shift = y.min % x.width
        if not shift:
            return return_(x)
        m1, m2 = x.min, x.max
        return return_(IntConstraint.range(x.width, m1>>shift, m2>>shift))

class IUshr(BaseOp):
    def __init__(self, parent, args):
        BaseOp.__init__(self, parent, args)
        self.rval = parent.makeVariable(args[0].type, origin=self)

    def propagateConstraints(self, x, y):
        M = 1<<x.width
        if y.min < y.max:
            intmax = (M//2)-1
            return return_(IntConstraint.range(x.width, min(x.min, 0), max(x.max, intmax)))
        shift = y.min % x.width
        if not shift:
            return return_(x)

        parts = [x.min, x.max]
        if x.min <= -1 <= x.max:
            parts.append(-1)
        if x.min <= 0 <= x.max:
            parts.append(0)
        parts = [p % M for p in parts]
        m1, m2 = min(parts), max(parts)

        return return_(IntConstraint.range(x.width, m1>>shift, m2>>shift))

#############################################################################################
exec_tts = excepttypes.Arithmetic,
class IDiv(BaseOp):
    def __init__(self, parent, args):
        super(IDiv, self).__init__(parent, args, makeException=True)
        self.rval = parent.makeVariable(args[0].type, origin=self)
        self.outExceptionCons = ObjectConstraint.fromTops(parent.env, [], exec_tts, nonnull=True)

    def propagateConstraints(self, x, y):
        excons = self.outExceptionCons if (y.min <= 0 <= y.max) else None
        if y.min == 0 == y.max:
            return throw(excons)

        # Calculate possible extremes for division, taking into account special case of intmin/-1
        intmin = -1<<(x.width - 1)
        xvals = set([x.min, x.max])
        yvals = set([y.min, y.max])

        for val in (intmin+1, 0):
            if x.min <= val <= x.max:
                xvals.add(val)
        for val in (-2,-1,1):
            if y.min <= val <= y.max:
                yvals.add(val)
        yvals.discard(0)

        vals = set()
        for xv, yv in itertools.product(xvals, yvals):
            if xv == intmin and yv == -1:
                vals.add(intmin)
            elif xv*yv < 0: # Unlike Python, Java rounds to 0 so opposite sign case must be handled specially
                vals.add(-(-xv//yv))
            else:
                vals.add(xv//yv)

        rvalcons = IntConstraint.range(x.width, min(vals), max(vals))
        return returnOrThrow(rvalcons, excons)

class IRem(BaseOp):
    def __init__(self, parent, args):
        super(IRem, self).__init__(parent, args, makeException=True)
        self.rval = parent.makeVariable(args[0].type, origin=self)
        self.outExceptionCons = ObjectConstraint.fromTops(parent.env, [], exec_tts, nonnull=True)

    def propagateConstraints(self, x, y):
        excons = self.outExceptionCons if (y.min <= 0 <= y.max) else None
        if y.min == 0 == y.max:
            return throw(excons)
        # only do an exact result if both values are constants, and otherwise
        # just approximate the range as -(y-1) to (y-1) (or 0 to y-1 if it's positive)
        if x.min == x.max and y.min == y.max:
            val = abs(x.min) % abs(y.min)
            val = val if x.min >= 0 else -val
            return return_(IntConstraint.range(x.width, val, val))

        mag = max(abs(y.min), abs(y.max)) - 1
        rmin = -min(mag, abs(x.min)) if x.min < 0 else 0
        rmax = min(mag, abs(x.max)) if x.max > 0 else 0

        rvalcons = IntConstraint.range(x.width, rmin, rmax)
        return returnOrThrow(rvalcons, excons)

###############################################################################
class ICmp(BaseOp):
    def __init__(self, parent, args):
        BaseOp.__init__(self, parent, args)
        self.rval = parent.makeVariable(ssa_types.SSA_INT, origin=self)

    def propagateConstraints(self, x, y):
        rvalcons = IntConstraint.range(32, -1, 1)
        return return_(rvalcons)



def get_attribute_raw(bytestream, ic_indices):
    name_ind, length = bytestream.get('>HL')

    # Hotspot does not actually check the attribute length of InnerClasses prior to 49.0
    # so this case requires special handling. We will keep the purported length of the
    # attribute so that it can be displayed in the disassembly. For InnerClass attributes
    # data is actually a (length, bytes) tuple, rather than storing the bytes directly
    if name_ind in ic_indices:
        count = bytestream.get('>H', peek=True)
        data = length, bytestream.getRaw(2+8*count)
    else:
        data = bytestream.getRaw(length)

    return name_ind,data

def get_attributes_raw(bytestream, ic_indices=()):
    attribute_count = bytestream.get('>H')
    return [get_attribute_raw(bytestream, ic_indices) for _ in range(attribute_count)]

def fixAttributeNames(attributes_raw, cpool):
    return [(cpool.getArgsCheck('Utf8', name_ind), data) for name_ind, data in attributes_raw]



ADD = 'add'
AND = 'and'
ANEWARRAY = 'anewarray'
ARRLEN = 'arrlen'
ARRLOAD = 'arrload'
ARRLOAD_OBJ = 'arrload_obj'
ARRSTORE = 'arrstore'
ARRSTORE_OBJ = 'arrstore_obj'
CHECKCAST = 'checkcast'
CONST = 'const'
CONSTNULL = 'constnull'
CONVERT = 'convert'
DIV = 'div'
DUP = 'dup'
DUP2 = 'dup2'
DUP2X1 = 'dup2x1'
DUP2X2 = 'dup2x2'
DUPX1 = 'dupx1'
DUPX2 = 'dupx2'
FCMP = 'fcmp'
GETFIELD = 'getfield'
GETSTATIC = 'getstatic'
GOTO = 'goto'
IF_A = 'if_a'
IF_ACMP = 'if_acmp'
IF_I = 'if_i'
IF_ICMP = 'if_icmp'
IINC = 'iinc'
INSTANCEOF = 'instanceof'
INVOKEDYNAMIC = 'invokedynamic'
INVOKEINIT = 'invokeinit'
INVOKEINTERFACE = 'invokeinterface'
INVOKESPECIAL = 'invokespecial'
INVOKESTATIC = 'invokestatic'
INVOKEVIRTUAL = 'invokevirtual'
JSR = 'jsr'
LCMP = 'lcmp'
LDC = 'ldc'
LOAD = 'load'
MONENTER = 'monenter'
MONEXIT = 'monexit'
MUL = 'mul'
MULTINEWARRAY = 'multinewarray'
NEG = 'neg'
NEW = 'new'
NEWARRAY = 'newarray'
NOP = 'nop'
OR = 'or'
POP = 'pop'
POP2 = 'pop2'
PUTFIELD = 'putfield'
PUTSTATIC = 'putstatic'
REM = 'rem'
RET = 'ret'
RETURN = 'return'
SHL = 'shl'
SHR = 'shr'
STORE = 'store'
SUB = 'sub'
SWAP = 'swap'
SWITCH = 'switch'
THROW = 'throw'
TRUNCATE = 'truncate'
USHR = 'ushr'
XOR = 'xor'



# double quote, backslash, and newlines are forbidden
ok_chars =  " !#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[]^_`abcdefghijklmnopqrstuvwxyz{|}~"
ok_chars = frozenset(ok_chars)

# these characters cannot use unicode escape codes due to the way Java escaping works
late_escape = {u'\u0009':r'\t', u'\u000a':r'\n', u'\u000d':r'\r', u'\u0022':r'\"', u'\u005c':r'\\'}

def escapeString(u):
    if set(u) <= ok_chars:
        return u

    escaped = []
    for c in u:
        if c in ok_chars:
            escaped.append(c)
        elif c in late_escape:
            escaped.append(late_escape[c])
        else:
            i = ord(c)
            if i <= 0xFFFF:
                escaped.append(r'\u{0:04x}'.format(i))
            else:
                i -= 0x10000
                high = 0xD800 + (i>>10)
                low = 0xDC00 + (i & 0x3FF)
                escaped.append(r'\u{0:04x}\u{1:04x}'.format(high,low))
    return ''.join(escaped)



import collections
import struct

# ConstantPool stores strings as strings or unicodes. They are automatically
# converted to and from modified Utf16 when reading and writing to binary

# Floats and Doubles are internally stored as integers with the same bit pattern
# Since using raw floats breaks equality testing for signed zeroes and NaNs
# cpool.getArgs/getArgsCheck will automatically convert them into Python floats

def decodeStr(s):
    return s.replace('\xc0\x80','\0').decode('utf8'),
def decodeFloat(i):
    return struct.unpack('>f', struct.pack('>i', i)) # Note: returns tuple
def decodeDouble(i):
    return struct.unpack('>d', struct.pack('>q', i))

cpoolInfo_t = collections.namedtuple('cpoolInfo_t',
                                     ['name','tag','recoverArgs'])

Utf8 = cpoolInfo_t('Utf8',1,
                  (lambda self,s:(s,)))

Class = cpoolInfo_t('Class',7,
                    (lambda self,n_id:self.getArgs(n_id)))

NameAndType = cpoolInfo_t('NameAndType',12,
                (lambda self,n,d:self.getArgs(n) + self.getArgs(d)))

Field = cpoolInfo_t('Field',9,
                (lambda self,c_id,nat_id:self.getArgs(c_id) + self.getArgs(nat_id)))

Method = cpoolInfo_t('Method',10,
                (lambda self,c_id,nat_id:self.getArgs(c_id) + self.getArgs(nat_id)))

InterfaceMethod = cpoolInfo_t('InterfaceMethod',11,
                (lambda self,c_id,nat_id:self.getArgs(c_id) + self.getArgs(nat_id)))

String = cpoolInfo_t('String',8,
                (lambda self,n_id:self.getArgs(n_id)))

Int = cpoolInfo_t('Int',3,
                  (lambda self,s:(s,)))

Long = cpoolInfo_t('Long',5,
                  (lambda self,s:(s,)))

Float = cpoolInfo_t('Float',4,
                  (lambda self,s:decodeFloat(s)))

Double = cpoolInfo_t('Double',6,
                  (lambda self,s:decodeDouble(s)))

MethodHandle = cpoolInfo_t('MethodHandle',15,
                (lambda self,t,n_id:(t,)+self.getArgs(n_id)))

MethodType = cpoolInfo_t('MethodType',16,
                (lambda self,n_id:self.getArgs(n_id)))

InvokeDynamic = cpoolInfo_t('InvokeDynamic',18,
                (lambda self,bs_id,nat_id:(bs_id,) + self.getArgs(nat_id)))

cpoolTypes = [Utf8, Class, NameAndType, Field, Method, InterfaceMethod,
              String, Int, Long, Float, Double,
              MethodHandle, MethodType, InvokeDynamic]
name2Type = {t.name:t for t in cpoolTypes}
tag2Type = {t.tag:t for t in cpoolTypes}

class ConstPool(object):
    def __init__(self, initialData=((None,None),)):
        self.pool = []
        self.reserved = set()
        self.available = set()

        for tag, val in initialData:
            if tag is None:
                self.addEmptySlot()
            else:
                t = tag2Type[tag]
                if t.name == 'Utf8':
                    val = decodeStr(*val)
                self.pool.append((t.name, val))

    def addEmptySlot(self):
        self.pool.append((None, None))

    def getArgs(self, i):
        if not (i >= 0 and i<len(self.pool)):
            raise IndexError('Constant pool index {} out of range'.format(i))
        if self.pool[i][0] is None:
            raise IndexError('Constant pool index {} invalid'.format(i))

        name, val = self.pool[i]
        t = name2Type[name]
        return t.recoverArgs(self, *val)

    def getArgsCheck(self, typen, index):
        # if (self.pool[index][0] != typen):
        #     raise KeyError('Constant pool index {} has incorrect type {}'.format(index, typen))
        val = self.getArgs(index)
        return val if len(val) > 1 else val[0]

    def getType(self, index): return self.pool[index][0]



from ..ssa import objtypes

from . import ast
from .stringescape import escapeString as escape

class Comments(object):
    def __init__(self):
        self.lines = []

    def add(self, s):
        self.lines.extend(s.strip('\n').split('\n'))

    def print_(self, printer, print_):
        return ''.join(map('// {}\n'.format, self.lines))

class MethodDef(object):
    def __init__(self, class_, flags, name, desc, retType, paramDecls, body):
        self.flagstr = flags + ' ' if flags else ''
        self.retType, self.paramDecls = retType, paramDecls
        self.body = body
        self.comments = Comments()
        self.triple = class_.name, name, desc
        self.throws = None

        if name == '<clinit>':
            self.isStaticInit, self.isConstructor = True, False
        elif name == '<init>':
            self.isStaticInit, self.isConstructor = False, True
            self.clsname = ast.TypeName(objtypes.TypeTT(class_.name, 0))
        else:
            self.isStaticInit, self.isConstructor = False, False

    def print_(self, printer, print_):
        header = print_(self.comments)
        argstr = ', '.join(print_(decl) for decl in self.paramDecls)
        if self.isStaticInit:
            header += 'static'
        elif self.isConstructor:
            name = print_(self.clsname).rpartition('.')[-1]
            header += '{}{}({})'.format(self.flagstr, name, argstr)
        else:
            name = printer.methodName(*self.triple)
            header += '{}{} {}({})'.format(self.flagstr, print_(self.retType), escape(name), argstr)

        if self.throws is not None:
            header += ' throws ' + print_(self.throws)

        if self.body is None:
            if 'abstract' not in self.flagstr and 'native' not in self.flagstr:
                # Create dummy body for decompiler error
                return header + ' {/*error*/throw null;}\n'
            return header + ';\n'
        else:
            return header + ' ' + print_(self.body)

    def tree(self, printer, tree):
        return {
            'triple': self.triple,
            'flags': self.flagstr.split(),
            'ret': tree(self.retType),
            'params': map(tree, self.paramDecls),
            'comments': self.comments.lines,
            'body': tree(self.body),
            'throws': tree(self.throws),
        }

class FieldDef(object):
    def __init__(self, flags, type_, class_, name, desc, expr=None):
        self.flagstr = flags + ' ' if flags else ''
        self.type_ = type_
        self.name = name
        self.expr = None if expr is None else ast.makeCastExpr(type_.tt, expr)
        self.triple = class_.name, name, desc

    def print_(self, printer, print_):
        name = escape(printer.fieldName(*self.triple))
        if self.expr is not None:
            return '{}{} {} = {};'.format(self.flagstr, print_(self.type_), name, print_(self.expr))
        return '{}{} {};'.format(self.flagstr, print_(self.type_), name)

    def tree(self, printer, tree):
        return {
            'triple': self.triple,
            'type': tree(self.type_),
            'flags': self.flagstr.split(),
            'expr': tree(self.expr),
        }

class ClassDef(object):
    def __init__(self, flags, isInterface, name, superc, interfaces, fields, methods):
        self.flagstr = flags + ' ' if flags else ''
        self.isInterface = isInterface
        self.name = ast.TypeName(objtypes.TypeTT(name,0))
        self.super = ast.TypeName(objtypes.TypeTT(superc,0)) if superc is not None else None
        self.interfaces = [ast.TypeName(objtypes.TypeTT(iname,0)) for iname in interfaces]
        self.fields = fields
        self.methods = methods
        if superc == 'java/lang/Object':
            self.super = None

    def print_(self, printer, print_):
        contents = ''
        if self.fields:
            contents = '\n'.join(print_(x) for x in self.fields)
        if self.methods:
            if contents:
                contents += '\n\n' # extra line to divide fields and methods
            contents += '\n\n'.join(print_(x) for x in self.methods)

        indented = ['    '+line for line in contents.splitlines()]
        name = print_(self.name).rpartition('.')[-1]
        defname = 'interface' if self.isInterface else 'class'
        header = '{}{} {}'.format(self.flagstr, defname, name)

        if self.super:
            header += ' extends ' + print_(self.super)
        if self.interfaces:
            if self.isInterface:
                assert self.super is None
                header += ' extends ' + ', '.join(print_(x) for x in self.interfaces)
            else:
                header += ' implements ' + ', '.join(print_(x) for x in self.interfaces)

        lines = [header + ' {'] + indented + ['}']
        return '\n'.join(lines) + '\n'

    # Experimental - don't use!
    def tree(self, printer, tree):
        return {
            'rawname': objtypes.className(self.name.tt),
            'name': tree(self.name),
            'super': tree(self.super),
            'flags': self.flagstr.split(),
            'isInterface': self.isInterface,
            'interfaces': map(tree, self.interfaces),
            'fields': map(tree, self.fields),
            'methods': map(tree, self.methods),
        }



from .verifier_types import T_ARRAY, T_BOOL, T_BYTE, T_CHAR, T_DOUBLE, T_FLOAT, T_INT, T_INVALID, T_LONG, T_OBJECT, T_SHORT, unSynthesizeType

_cat2tops = T_LONG, T_DOUBLE

def parseFieldDescriptors(desc_str, unsynthesize=True):
    baseTypes = {'B':T_BYTE, 'C':T_CHAR, 'D':T_DOUBLE, 'F':T_FLOAT,
                 'I':T_INT, 'J':T_LONG, 'S':T_SHORT, 'Z':T_BOOL}

    fields = []
    while desc_str:
        oldlen = len(desc_str)
        desc_str = desc_str.lstrip('[')
        dim = oldlen - len(desc_str)
        if dim > 255:
            raise ValueError('Dimension {} > 255 in descriptor'.format(dim))
        if not desc_str:
            raise ValueError('Descriptor contains [s at end of string')

        if desc_str[0] == 'L':
            end = desc_str.find(';')
            if end == -1:
                raise ValueError('Unmatched L in descriptor')

            name = desc_str[1:end]
            desc_str = desc_str[end+1:]
            baset = T_OBJECT(name)
        else:
            if desc_str[0] not in baseTypes:
                raise ValueError('Unrecognized code {} in descriptor'.format(desc_str[0]))
            baset = baseTypes[desc_str[0]]
            desc_str = desc_str[1:]

        if dim:
            # Hotspot considers byte[] and bool[] identical for type checking purposes
            if unsynthesize and baset == T_BOOL:
                baset = T_BYTE
            baset = T_ARRAY(baset, dim)
        elif unsynthesize:
            # synthetics are only meaningful as basetype of an array
            # if they are by themselves, convert to int.
            baset = unSynthesizeType(baset)

        fields.append(baset)
        if baset in _cat2tops:
            fields.append(T_INVALID)
    return fields

# get a single descriptor
def parseFieldDescriptor(desc_str, unsynthesize=True):
    rval = parseFieldDescriptors(desc_str, unsynthesize)

    cat = 2 if (rval and rval[0] in _cat2tops) else 1
    if len(rval) != cat:
        raise ValueError('Incorrect number of fields in descriptor, expected {} but found {}'.format(cat, len(rval)))
    return rval

# Parse a string to get a Java Method Descriptor
def parseMethodDescriptor(desc_str, unsynthesize=True):
    if not desc_str.startswith('('):
        raise ValueError('Method descriptor does not start with (')

    # we need to split apart the argument list and return value
    # this is greatly complicated by the fact that ) is a legal
    # character that can appear in class names

    lp_pos = desc_str.rfind(')') # this case will work if return type is not an object
    if desc_str.endswith(';'):
        lbound = max(desc_str.rfind(';', 1, -1), 1)
        lp_pos = desc_str.find(')', lbound, -1)
    if lp_pos < 0 or desc_str[lp_pos] != ')':
        raise ValueError('Unable to split method descriptor into arguments and return type')

    arg_str = desc_str[1:lp_pos]
    rval_str = desc_str[lp_pos+1:]

    args = parseFieldDescriptors(arg_str, unsynthesize)
    rval = [] if rval_str == 'V' else parseFieldDescriptor(rval_str, unsynthesize)
    return args, rval

# Adds self argument for nonstatic. Constructors must be handled seperately
def parseUnboundMethodDescriptor(desc_str, target, isstatic):
    args, rval = parseMethodDescriptor(desc_str)
    if not isstatic:
        args = [T_OBJECT(target)] + args
    return args, rval



from .attributes_raw import fixAttributeNames

class Field(object):
    flagVals = {'PUBLIC':0x0001,
                'PRIVATE':0x0002,
                'PROTECTED':0x0004,
                'STATIC':0x0008,
                'FINAL':0x0010,
                'VOLATILE':0x0040,
                'TRANSIENT':0x0080,
                'SYNTHETIC':0x1000,
                'ENUM':0x4000,
                }

    def __init__(self, data, classFile, keepRaw):
        self.class_ = classFile
        cpool = self.class_.cpool

        flags, name_id, desc_id, attributes_raw = data

        self.name = cpool.getArgsCheck('Utf8', name_id)
        self.descriptor = cpool.getArgsCheck('Utf8', desc_id)
        self.attributes = fixAttributeNames(attributes_raw, cpool)

        self.flags = set(name for name,mask in Field.flagVals.items() if (mask & flags))
        self.static = 'STATIC' in self.flags
        if keepRaw:
            self.attributes_raw = attributes_raw
            self.name_id, self.desc_id = name_id, desc_id



import itertools

from .. import objtypes
from ..mixin import ValueType

array_supers = 'java/lang/Object','java/lang/Cloneable','java/io/Serializable'
obj_fset = frozenset([objtypes.ObjectTT])

def isAnySubtype(env, x, seq):
    return any(objtypes.isSubtype(env,x,y) for y in seq)

class TypeConstraint(ValueType):
    __slots__ = "env supers exact isBot".split()
    def __init__(self, env, supers, exact):
        self.env, self.supers, self.exact = env, frozenset(supers), frozenset(exact)
        self.isBot = objtypes.ObjectTT in supers

        temp = self.supers | self.exact
        assert objtypes.NullTT not in temp
        assert all(objtypes.isBaseTClass(tt) for tt in supers)
        assert all(objtypes.dim(tt) < 999 for tt in exact)

    def _key(self): return self.supers, self.exact
    def __nonzero__(self): return bool(self.supers or self.exact)

    def getSingleTType(self):
        # comSuper doesn't care about order so we can freely pass in nondeterministic order
        return objtypes.commonSupertype(self.env, list(self.supers) + list(self.exact))

    def isBoolOrByteArray(self):
        if self.supers or len(self.exact) != 2:
            return False
        tt1, tt2 = self.exact
        bases = objtypes.baset(tt1), objtypes.baset(tt2)
        return objtypes.dim(tt1) == objtypes.dim(tt2) and sorted(bases) == [objtypes.baset(objtypes.BoolTT), objtypes.baset(objtypes.ByteTT)]

    @staticmethod
    def reduce(env, supers, exact):
        newsupers = []
        for x in supers:
            if not isAnySubtype(env, x, newsupers):
                newsupers = [y for y in newsupers if not objtypes.isSubtype(env, y, x)]
                newsupers.append(x)

        newexact = [x for x in exact if not isAnySubtype(env, x, newsupers)]
        return TypeConstraint(env, newsupers, newexact)

    def join(*cons):
        assert len(set(map(type, cons))) == 1
        env = cons[0].env

        # optimize for the common case of joining with itself or with bot
        cons = set(c for c in cons if not c.isBot)
        if not cons:
            return TypeConstraint(env, obj_fset, [])
        elif len(cons) == 1:
            return cons.pop()
        assert(len(cons) == 2) # joining more than 2 not currently supported

        supers_l, exact_l = zip(*(c._key() for c in cons))

        newsupers = set()
        for t1,t2 in itertools.product(*supers_l):
            if objtypes.isSubtype(env, t1, t2):
                newsupers.add(t1)
            elif objtypes.isSubtype(env, t2, t1):
                newsupers.add(t2)
            else: # TODO: need to add special handling for interfaces here
                pass

        newexact = frozenset.union(*exact_l)
        for c in cons:
            newexact = [x for x in newexact if x in c.exact or isAnySubtype(env, x, c.supers)]
        return TypeConstraint.reduce(env, newsupers, newexact)

    def meet(*cons):
        supers = frozenset.union(*(c.supers for c in cons))
        exact = frozenset.union(*(c.exact for c in cons))
        return TypeConstraint.reduce(cons[0].env, supers, exact)

class ObjectConstraint(ValueType):
    __slots__ = "null types isBot".split()
    def __init__(self, null, types):
        self.null, self.types = null, types
        self.isBot = null and types.isBot

    @staticmethod
    def constNull(env):
        return ObjectConstraint(True, TypeConstraint(env, [], []))

    @staticmethod
    def fromTops(env, supers, exact, nonnull=False):
        types = TypeConstraint(env, supers, exact)
        if nonnull and not types:
            return None
        return ObjectConstraint(not nonnull, types)

    def _key(self): return self.null, self.types

    def isConstNull(self): return self.null and not self.types

    def getSingleTType(self):
        return self.types.getSingleTType() if self.types else objtypes.NullTT

    def join(*cons):
        null = all(c.null for c in cons)
        types = TypeConstraint.join(*(c.types for c in cons))
        if not null and not types:
            return None

        res = ObjectConstraint(null, types)
        return cons[0] if cons[0] == res else res

    def meet(*cons):
        null = any(c.null for c in cons)
        types = TypeConstraint.meet(*(c.types for c in cons))
        return  ObjectConstraint(null, types)

    def __str__(self):   # pragma: no cover
        if not self.types:
            return 'Obj(null)'
        return 'Obj({}, {}, {})'.format(self.null, sorted(self.types.supers), sorted(self.types.exact))
    __repr__ = __str__



from .. import opnames as vops
from ..verifier.descriptors import parseFieldDescriptor, parseMethodDescriptor

from . import objtypes, ssa_jumps, ssa_ops, subproc
from .ssa_types import SSA_DOUBLE, SSA_FLOAT, SSA_INT, SSA_LONG, SSA_OBJECT, slots_t


_charToSSAType = {'D':SSA_DOUBLE, 'F':SSA_FLOAT, 'I':SSA_INT, 'J':SSA_LONG,
                'B':SSA_INT, 'C':SSA_INT, 'S':SSA_INT}
def getCategory(c): return 2 if c in 'JD' else 1

class ResultDict(object):
    def __init__(self, line=None, jump=None, newstack=None, newlocals=None):
        self.line = line
        self.jump = jump
        self.newstack = newstack
        self.newlocals = newlocals

##############################################################################
def makeConstVar(parent, type_, val):
    var = parent.makeVariable(type_)
    var.const = val
    return var

def parseArrOrClassName(desc):
    # Accept either a class or array descriptor or a raw class name.
    if desc.startswith('[') or desc.endswith(';'):
        vtypes = parseFieldDescriptor(desc, unsynthesize=False)
        tt = objtypes.verifierToSynthetic(vtypes[0])
    else:
        tt = objtypes.TypeTT(desc, 0)
    return tt

def _floatOrIntMath(fop, iop):
    def math1(maker, input_, iNode):
        cat = getCategory(iNode.instruction[1])
        isfloat = (iNode.instruction[1] in 'DF')
        op = fop if isfloat else iop

        args = input_.stack[-cat*2::cat]
        line = op(maker.parent, args)

        newstack = input_.stack[:-2*cat] + [line.rval] + [None]*(cat-1)
        return ResultDict(line=line, newstack=newstack)
    return math1

def _intMath(op, isShift):
    def math2(maker, input_, iNode):
        cat = getCategory(iNode.instruction[1])
        # some ops (i.e. shifts) always take int as second argument
        size = cat+1 if isShift else cat+cat
        args = input_.stack[-size::cat]
        line = op(maker.parent, args)
        newstack = input_.stack[:-size] + [line.rval] + [None]*(cat-1)
        return ResultDict(line=line, newstack=newstack)
    return math2
##############################################################################

def _anewarray(maker, input_, iNode):
    name = maker.parent.getConstPoolArgs(iNode.instruction[1])[0]
    tt = parseArrOrClassName(name)
    line = ssa_ops.NewArray(maker.parent, input_.stack[-1], tt)
    newstack = input_.stack[:-1] + [line.rval]
    return ResultDict(line=line, newstack=newstack)

def _arrlen(maker, input_, iNode):
    line = ssa_ops.ArrLength(maker.parent, input_.stack[-1:])
    newstack = input_.stack[:-1] + [line.rval]
    return ResultDict(line=line, newstack=newstack)

def _arrload(maker, input_, iNode):
    type_ = _charToSSAType[iNode.instruction[1]]
    cat = getCategory(iNode.instruction[1])

    line = ssa_ops.ArrLoad(maker.parent, input_.stack[-2:], type_)
    newstack = input_.stack[:-2] + [line.rval] + [None]*(cat-1)
    return ResultDict(line=line, newstack=newstack)

def _arrload_obj(maker, input_, iNode):
    line = ssa_ops.ArrLoad(maker.parent, input_.stack[-2:], SSA_OBJECT)
    newstack = input_.stack[:-2] + [line.rval]
    return ResultDict(line=line, newstack=newstack)

def _arrstore(maker, input_, iNode):
    if getCategory(iNode.instruction[1]) > 1:
        newstack, args = input_.stack[:-4], input_.stack[-4:-1]
        arr_vt, ind_vt = iNode.state.stack[-4:-2]
    else:
        newstack, args = input_.stack[:-3], input_.stack[-3:]
        arr_vt, ind_vt = iNode.state.stack[-3:-1]
    line = ssa_ops.ArrStore(maker.parent, args)

    # Check if we can prune the exception early because the
    # array size and index are known constants
    if arr_vt.const is not None and ind_vt.const is not None:
        if 0 <= ind_vt.const < arr_vt.const:
            line.outException = None
    return ResultDict(line=line, newstack=newstack)

def _arrstore_obj(maker, input_, iNode):
    line = ssa_ops.ArrStore(maker.parent, input_.stack[-3:])
    newstack = input_.stack[:-3]
    return ResultDict(line=line, newstack=newstack)

def _checkcast(maker, input_, iNode):
    index = iNode.instruction[1]
    desc = maker.parent.getConstPoolArgs(index)[0]
    tt = parseArrOrClassName(desc)
    line = ssa_ops.CheckCast(maker.parent, tt, input_.stack[-1:])
    return ResultDict(line=line)

def _const(maker, input_, iNode):
    ctype, val = iNode.instruction[1:]
    cat = getCategory(ctype)
    type_ = _charToSSAType[ctype]
    var = makeConstVar(maker.parent, type_, val)
    newstack = input_.stack + [var] + [None]*(cat-1)
    return ResultDict(newstack=newstack)

def _constnull(maker, input_, iNode):
    var = makeConstVar(maker.parent, SSA_OBJECT, 'null')
    var.decltype = objtypes.NullTT
    newstack = input_.stack + [var]
    return ResultDict(newstack=newstack)

def _convert(maker, input_, iNode):
    src_c, dest_c = iNode.instruction[1:]
    src_cat, dest_cat = getCategory(src_c), getCategory(dest_c)

    stack, arg =  input_.stack[:-src_cat], input_.stack[-src_cat]
    line = ssa_ops.Convert(maker.parent, arg, _charToSSAType[src_c], _charToSSAType[dest_c])

    newstack = stack + [line.rval] + [None]*(dest_cat-1)
    return ResultDict(line=line, newstack=newstack)

def _fcmp(maker, input_, iNode):
    op, c, NaN_val = iNode.instruction
    cat = getCategory(c)

    args = input_.stack[-cat*2::cat]
    line = ssa_ops.FCmp(maker.parent, args, NaN_val)
    newstack = input_.stack[:-cat*2] + [line.rval]
    return ResultDict(line=line, newstack=newstack)

def _field_access(maker, input_, iNode):
    index = iNode.instruction[1]
    target, name, desc = maker.parent.getConstPoolArgs(index)
    cat = len(parseFieldDescriptor(desc))

    argcnt = cat if 'put' in iNode.instruction[0] else 0
    if not 'static' in iNode.instruction[0]:
        argcnt += 1
    splitInd = len(input_.stack) - argcnt

    args = [x for x in input_.stack[splitInd:] if x is not None]
    line = ssa_ops.FieldAccess(maker.parent, iNode.instruction, (target, name, desc), args=args)
    newstack = input_.stack[:splitInd] + line.returned
    return ResultDict(line=line, newstack=newstack)

def _goto(maker, input_, iNode):
    jump = ssa_jumps.Goto(maker.parent, maker.blockd[iNode.successors[0]])
    return ResultDict(jump=jump)

def _if_a(maker, input_, iNode):
    null = makeConstVar(maker.parent, SSA_OBJECT, 'null')
    null.decltype = objtypes.NullTT
    jump = ssa_jumps.If(maker.parent, iNode.instruction[1], map(maker.blockd.get, iNode.successors), (input_.stack[-1], null))
    newstack = input_.stack[:-1]
    return ResultDict(jump=jump, newstack=newstack)

def _if_i(maker, input_, iNode):
    zero = makeConstVar(maker.parent, SSA_INT, 0)
    jump = ssa_jumps.If(maker.parent, iNode.instruction[1], map(maker.blockd.get, iNode.successors), (input_.stack[-1], zero))
    newstack = input_.stack[:-1]
    return ResultDict(jump=jump, newstack=newstack)

def _if_cmp(maker, input_, iNode):
    jump = ssa_jumps.If(maker.parent, iNode.instruction[1], map(maker.blockd.get, iNode.successors), input_.stack[-2:])
    newstack = input_.stack[:-2]
    return ResultDict(jump=jump, newstack=newstack)

def _iinc(maker, input_, iNode):
    _, index, amount = iNode.instruction

    oldval = input_.locals[index]
    constval = makeConstVar(maker.parent, SSA_INT, amount)
    line = ssa_ops.IAdd(maker.parent, (oldval, constval))

    newlocals = input_.locals.copy()
    newlocals[index] = line.rval
    return ResultDict(line=line, newlocals=newlocals)

def _instanceof(maker, input_, iNode):
    index = iNode.instruction[1]
    desc = maker.parent.getConstPoolArgs(index)[0]
    tt = parseArrOrClassName(desc)
    line = ssa_ops.InstanceOf(maker.parent, tt, input_.stack[-1:])
    newstack = input_.stack[:-1] + [line.rval]
    return ResultDict(line=line, newstack=newstack)

def _invoke(maker, input_, iNode):
    index = iNode.instruction[1]
    target, name, desc = maker.parent.getConstPoolArgs(index)
    target_tt = parseArrOrClassName(target)

    argcnt = len(parseMethodDescriptor(desc)[0])
    if not 'static' in iNode.instruction[0]:
        argcnt += 1
    splitInd = len(input_.stack) - argcnt

    # If we are an initializer, store a copy of the uninitialized verifier type so the Java decompiler can patch things up later
    isThisCtor = iNode.isThisCtor if iNode.op == vops.INVOKEINIT else False

    args = [x for x in input_.stack[splitInd:] if x is not None]
    line = ssa_ops.Invoke(maker.parent, iNode.instruction, (target, name, desc),
        args=args, isThisCtor=isThisCtor, target_tt=target_tt)
    newstack = input_.stack[:splitInd] + line.returned
    return ResultDict(line=line, newstack=newstack)

def _invoke_dynamic(maker, input_, iNode):
    index = iNode.instruction[1]
    desc = maker.parent.getConstPoolArgs(index)[2]
    argcnt = len(parseMethodDescriptor(desc)[0])
    splitInd = len(input_.stack) - argcnt

    args = [x for x in input_.stack[splitInd:] if x is not None]
    line = ssa_ops.InvokeDynamic(maker.parent, desc, args)
    newstack = input_.stack[:splitInd] + line.returned
    return ResultDict(line=line, newstack=newstack)

def _jsr(maker, input_, iNode):
    newstack = input_.stack + [None]
    if iNode.returnedFrom is None:
        jump = ssa_jumps.Goto(maker.parent, maker.blockd[iNode.successors[0]])
        return ResultDict(newstack=newstack, jump=jump)

    # create output variables from callop to represent vars received from ret.
    # We can use {} for initMap since there will never be unintialized types here
    retnode = maker.iNodeD[iNode.returnedFrom]
    stack = [maker.parent.makeVarFromVtype(vt, {}) for vt in retnode.out_state.stack]
    newlocals = dict(enumerate(maker.parent.makeVarFromVtype(vt, {}) for vt in retnode.out_state.locals))
    newlocals = maker.pruneUnused(retnode.key, newlocals)
    out_slots = slots_t(locals=newlocals, stack=stack)

    # Simply store the data for now and fix things up once all the blocks are created
    jump = subproc.ProcCallOp(maker.blockd[iNode.successors[0]], maker.blockd[iNode.next_instruction], input_, out_slots)
    return ResultDict(jump=jump, newstack=newstack)

def _lcmp(maker, input_, iNode):
    args = input_.stack[-4::2]
    line = ssa_ops.ICmp(maker.parent, args)
    newstack = input_.stack[:-4] + [line.rval]
    return ResultDict(line=line, newstack=newstack)

def _ldc(maker, input_, iNode):
    index, cat = iNode.instruction[1:]
    entry_type = maker.parent.getConstPoolType(index)
    args = maker.parent.getConstPoolArgs(index)

    var = None
    if entry_type == 'String':
        var = makeConstVar(maker.parent, SSA_OBJECT, args[0])
        var.decltype = objtypes.StringTT
    elif entry_type == 'Int':
        var = makeConstVar(maker.parent, SSA_INT, args[0])
    elif entry_type == 'Long':
        var = makeConstVar(maker.parent, SSA_LONG, args[0])
    elif entry_type == 'Float':
        var = makeConstVar(maker.parent, SSA_FLOAT, args[0])
    elif entry_type == 'Double':
        var = makeConstVar(maker.parent, SSA_DOUBLE, args[0])
    elif entry_type == 'Class':
        var = makeConstVar(maker.parent, SSA_OBJECT, parseArrOrClassName(args[0]))
        var.decltype = objtypes.ClassTT
    # Todo - handle MethodTypes and MethodHandles?

    assert var
    newstack = input_.stack + [var] + [None]*(cat-1)
    return ResultDict(newstack=newstack)

def _load(maker, input_, iNode):
    cat = getCategory(iNode.instruction[1])
    index = iNode.instruction[2]
    newstack = input_.stack + [input_.locals[index]] + [None]*(cat-1)
    return ResultDict(newstack=newstack)

def _monitor(maker, input_, iNode):
    isExit = 'exit' in iNode.instruction[0]
    line = ssa_ops.Monitor(maker.parent, input_.stack[-1:], isExit)
    newstack = input_.stack[:-1]
    return ResultDict(line=line, newstack=newstack)

def _multinewarray(maker, input_, iNode):
    op, index, dim = iNode.instruction
    name = maker.parent.getConstPoolArgs(index)[0]
    tt = parseArrOrClassName(name)
    assert objtypes.dim(tt) >= dim

    line = ssa_ops.MultiNewArray(maker.parent, input_.stack[-dim:], tt)
    newstack = input_.stack[:-dim] + [line.rval]
    return ResultDict(line=line, newstack=newstack)

def _neg(maker, input_, iNode):
    cat = getCategory(iNode.instruction[1])
    arg = input_.stack[-cat:][0]

    if (iNode.instruction[1] in 'DF'):
        line = ssa_ops.FNeg(maker.parent, [arg])
    else: # for integers, we can just write -x as 0 - x
        zero = makeConstVar(maker.parent, arg.type, 0)
        line = ssa_ops.ISub(maker.parent, [zero,arg])

    newstack = input_.stack[:-cat] + [line.rval] + [None]*(cat-1)
    return ResultDict(line=line, newstack=newstack)

def _new(maker, input_, iNode):
    index = iNode.instruction[1]
    classname = maker.parent.getConstPoolArgs(index)[0]
    if classname.endswith(';'):
        classname = classname[1:-1]

    line = ssa_ops.New(maker.parent, classname, iNode.key)
    newstack = input_.stack + [line.rval]
    return ResultDict(line=line, newstack=newstack)

def _newarray(maker, input_, iNode):
    vtypes = parseFieldDescriptor(iNode.instruction[1], unsynthesize=False)
    tt = objtypes.verifierToSynthetic(vtypes[0])

    line = ssa_ops.NewArray(maker.parent, input_.stack[-1], tt)
    newstack = input_.stack[:-1] + [line.rval]
    return ResultDict(line=line, newstack=newstack)

def _nop(maker, input_, iNode):
    return ResultDict()

def _ret(maker, input_, iNode):
    jump = subproc.DummyRet(input_, maker.blockd[iNode.jsrTarget])
    return ResultDict(jump=jump)

def _return(maker, input_, iNode):
    # Our special return block expects only the return values on the stack
    rtype = iNode.instruction[1]
    if rtype is None:
        newstack = []
    else:
        newstack = input_.stack[-getCategory(rtype):]

    # TODO: enable once structuring is smarter
    # if not maker.hasmonenter:
    #     jump = ssa_jumps.Goto(maker.parent, maker.returnBlock)
    #     return ResultDict(jump=jump, newstack=newstack)

    line = ssa_ops.TryReturn(maker.parent)
    return ResultDict(line=line, newstack=newstack)

def _store(maker, input_, iNode):
    cat = getCategory(iNode.instruction[1])
    index = iNode.instruction[2]

    newlocals = input_.locals.copy()
    newlocals[index] = input_.stack[-cat]
    newstack = input_.stack[:-cat]
    return ResultDict(newstack=newstack, newlocals=newlocals)

def _switch(maker, input_, iNode):
    default, raw_table = iNode.instruction[1:3]
    table = [(k, maker.blockd[v]) for k,v in raw_table]
    jump = ssa_jumps.Switch(maker.parent, maker.blockd[default], table, input_.stack[-1:])
    newstack = input_.stack[:-1]
    return ResultDict(jump=jump, newstack=newstack)

def _throw(maker, input_, iNode):
    line = ssa_ops.Throw(maker.parent, input_.stack[-1:])
    return ResultDict(line=line, newstack=[])

def _truncate(maker, input_, iNode):
    dest_c = iNode.instruction[1]
    signed, width = {'B':(True, 8), 'C':(False, 16), 'S':(True, 16)}[dest_c]

    line = ssa_ops.Truncate(maker.parent, input_.stack[-1], signed=signed, width=width)
    newstack = input_.stack[:-1] + [line.rval]
    return ResultDict(line=line, newstack=newstack)

def genericStackUpdate(maker, input_, iNode):
    n = iNode.pop_amount
    stack = input_.stack
    stack, popped = stack[:-n], stack[-n:]

    for i in iNode.stack_code:
        stack.append(popped[i])
    return ResultDict(newstack=stack)

instructionHandlers = {
                        vops.ADD: _floatOrIntMath(ssa_ops.FAdd, ssa_ops.IAdd),
                        vops.AND: _intMath(ssa_ops.IAnd, isShift=False),
                        vops.ANEWARRAY: _anewarray,
                        vops.ARRLEN: _arrlen,
                        vops.ARRLOAD: _arrload,
                        vops.ARRLOAD_OBJ: _arrload_obj,
                        vops.ARRSTORE: _arrstore,
                        vops.ARRSTORE_OBJ: _arrstore_obj,
                        vops.CHECKCAST: _checkcast,
                        vops.CONST: _const,
                        vops.CONSTNULL: _constnull,
                        vops.CONVERT: _convert,
                        vops.DIV: _floatOrIntMath(ssa_ops.FDiv, ssa_ops.IDiv),
                        vops.FCMP: _fcmp,
                        vops.GETSTATIC: _field_access,
                        vops.GETFIELD: _field_access,
                        vops.GOTO: _goto,
                        vops.IF_A: _if_a,
                        vops.IF_ACMP: _if_cmp, # cmp works on ints or objs
                        vops.IF_I: _if_i,
                        vops.IF_ICMP: _if_cmp,
                        vops.IINC: _iinc,
                        vops.INSTANCEOF: _instanceof,
                        vops.INVOKEINIT: _invoke,
                        vops.INVOKEINTERFACE: _invoke,
                        vops.INVOKESPECIAL: _invoke,
                        vops.INVOKESTATIC: _invoke,
                        vops.INVOKEVIRTUAL: _invoke,
                        vops.INVOKEDYNAMIC: _invoke_dynamic,
                        vops.JSR: _jsr,
                        vops.LCMP: _lcmp,
                        vops.LDC: _ldc,
                        vops.LOAD: _load,
                        vops.MONENTER: _monitor,
                        vops.MONEXIT: _monitor,
                        vops.MULTINEWARRAY: _multinewarray,
                        vops.MUL: _floatOrIntMath(ssa_ops.FMul, ssa_ops.IMul),
                        vops.NEG: _neg,
                        vops.NEW: _new,
                        vops.NEWARRAY: _newarray,
                        vops.NOP: _nop,
                        vops.OR: _intMath(ssa_ops.IOr, isShift=False),
                        vops.PUTSTATIC: _field_access,
                        vops.PUTFIELD: _field_access,
                        vops.REM: _floatOrIntMath(ssa_ops.FRem, ssa_ops.IRem),
                        vops.RET: _ret,
                        vops.RETURN: _return,
                        vops.SHL: _intMath(ssa_ops.IShl, isShift=True),
                        vops.SHR: _intMath(ssa_ops.IShr, isShift=True),
                        vops.STORE: _store,
                        vops.SUB: _floatOrIntMath(ssa_ops.FSub, ssa_ops.ISub),
                        vops.SWITCH: _switch,
                        vops.THROW: _throw,
                        vops.TRUNCATE: _truncate,
                        vops.USHR: _intMath(ssa_ops.IUshr, isShift=True),
                        vops.XOR: _intMath(ssa_ops.IXor, isShift=False),

                        vops.SWAP: genericStackUpdate,
                        vops.POP: genericStackUpdate,
                        vops.POP2: genericStackUpdate,
                        vops.DUP: genericStackUpdate,
                        vops.DUPX1: genericStackUpdate,
                        vops.DUPX2: genericStackUpdate,
                        vops.DUP2: genericStackUpdate,
                        vops.DUP2X1: genericStackUpdate,
                        vops.DUP2X2: genericStackUpdate,
                        }



import struct


class TruncatedStreamError(EOFError):
    pass


class Reader(object):
    __slots__ = ['d', 'off']
    def __init__(self, data, off=0):
        self.d = data
        self.off = off

    def done(self): return self.off >= len(self.d)
    def copy(self): return Reader(self.d, self.off)

    def u8(self): return self.get('>B')
    def s8(self): return self.get('>b')
    def u16(self): return self.get('>H')
    def s16(self): return self.get('>h')
    def u32(self): return self.get('>I')
    def s32(self): return self.get('>i')
    def u64(self): return self.get('>Q')

    # binUnpacker functions
    def get(self, fmt, forceTuple=False, peek=False):
        size = struct.calcsize(fmt)
        if self.size() < size:
            raise TruncatedStreamError()

        val = struct.unpack_from(fmt, self.d, self.off)

        if not peek:
            self.off += size
        if not forceTuple and len(val) == 1:
            val = val[0]
        return val

    def getRaw(self, num):
        if self.size() < num:
            raise TruncatedStreamError()
        val = self.d[self.off:self.off+num]
        self.off += num
        return val

    def size(self):
        return len(self.d) - self.off



import collections
import operator

from .. import opnames as vops
from ..verifier import verifier_types

from . import ssa_jumps, ssa_ops, subproc
from .blockmakerfuncs import ResultDict, instructionHandlers
from .ssa_types import BasicBlock, SSA_OBJECT, slots_t

def toBits(x): return [i for i in range(x.bit_length()) if x & (1 << i)]

# keys for special blocks created at the cfg entry and exit. Negative keys ensures they don't collide
ENTRY_KEY, RETURN_KEY, RETHROW_KEY = -1, -2, -3

def getUsedLocals(iNodes, iNodeD, exceptions):
    # For every instruction, find which locals at that point may be used in the future
    except_ranges = [(h, [node.key for node in iNodes if s <= node.key < e]) for s, e, h, i in exceptions]

    old = collections.defaultdict(int)
    while 1:
        data = old.copy()
        # Do one iteration
        for node in reversed(iNodes):
            used = reduce(operator.__or__, (data[key] for key in node.successors), 0)

            if node.instruction[0] == vops.LOAD:
                used |= 1 << node.instruction[2]
            elif node.instruction[0] == vops.IINC:
                used |= 1 << node.instruction[1]
            elif node.instruction[0] == vops.STORE:
                bits = 3 if node.instruction[1] in 'JD' else 1
                used &= ~(bits << node.instruction[2])
            elif node.instruction[0] == vops.RET:
                # If local is not in mask, it will use the value from the jsr instead of the ret
                mask = sum(1<<i for i in node.out_state.copy().maskFor(node.jsrTarget))
                used &= mask
            elif node.instruction[0] == vops.JSR and node.returnedFrom is not None:
                retnode = iNodeD[node.returnedFrom]
                assert node.successors == (retnode.jsrTarget,)
                mask = sum(1<<i for i in iNodeD[node.returnedFrom].out_state.copy().maskFor(retnode.jsrTarget))

                assert node.next_instruction is not None
                used |= (data[node.next_instruction] & ~mask)
            data[node.key] |= used

        for hkey, region in except_ranges:
            if data[hkey] != old[hkey]:
                for key in region:
                    data[key] |= data[hkey]

        if data == old:
            break
        old = data
    # for entry point, every program argument is marked used so we can preserve input arguments for later
    old[ENTRY_KEY] = (1 << len(iNodeD[0].state.locals)) - 1
    return old

def slotsRvals(inslots):
    stack = [(None if phi is None else phi.rval) for phi in inslots.stack]
    newlocals = {i: phi.rval for i, phi in inslots.locals.items() if phi is not None}
    return slots_t(stack=stack, locals=newlocals)

_jump_instrs = frozenset([vops.GOTO, vops.IF_A, vops.IF_ACMP, vops.IF_I, vops.IF_ICMP, vops.JSR, vops.SWITCH])
class BlockMaker(object):
    def __init__(self, parent, iNodes, inputTypes, returnTypes, except_raw, opts):
        self.parent = parent
        self.blocks = []
        self.blockd = {}

        self.iNodes = [n for n in iNodes if n.visited]
        self.iNodeD = {n.key: n for n in self.iNodes}
        exceptions = [eh for eh in except_raw if eh.handler in self.iNodeD]

        # Calculate which locals are actually live at any point
        self.used_locals = getUsedLocals(self.iNodes, self.iNodeD, exceptions)

        # create map of uninitialized -> initialized types so we can convert them
        self.initMap = {}
        for node in self.iNodes:
            if node.op == vops.NEW:
                self.initMap[node.stack_push[0]] = node.target_type
        self.initMap[verifier_types.T_UNINIT_THIS] = verifier_types.T_OBJECT(parent.class_.name)
        self.hasmonenter = any(node.instruction[0] == vops.MONENTER for node in self.iNodes)

        self.entryBlock = self.makeBlockWithInslots(ENTRY_KEY, newlocals=inputTypes, stack=[])
        self.returnBlock = self.makeBlockWithInslots(RETURN_KEY, newlocals=[], stack=returnTypes)
        self.returnBlock.jump = ssa_jumps.Return(self, [phi.rval for phi in self.returnBlock.phis])
        self.rethrowBlock = self.makeBlockWithInslots(RETHROW_KEY, newlocals=[], stack=[verifier_types.THROWABLE_INFO])
        self.rethrowBlock.jump = ssa_jumps.Rethrow(self, [phi.rval for phi in self.rethrowBlock.phis])

        # for ssagraph to copy
        self.inputArgs = slotsRvals(self.entryBlock.inslots).localsAsList
        self.entryBlock.phis = []

        # We need to create stub blocks for every jump target so we can add them as successors during creation
        jump_targets = [eh.handler for eh in exceptions]
        for node in self.iNodes:
            if node.instruction[0] in _jump_instrs:
                jump_targets += node.successors
            # add jsr fallthroughs too
            if node.instruction[0] == vops.JSR and node.returnedFrom is not None:
                jump_targets.append(node.next_instruction)

        # for simplicity, keep jsr stuff in individual instruction blocks.
        # Note that subproc.py will need to be modified if this is changed
        for node in self.iNodes:
            if node.instruction[0] in (vops.JSR, vops.RET):
                jump_targets.append(node.key)
        for key in jump_targets:
            if key not in self.blockd: # jump_targets may have duplicates
                self.makeBlock(key)

        self.exceptionhandlers = []
        for (start, end, handler, index) in exceptions:
            catchtype = parent.getConstPoolArgs(index)[0] if index else 'java/lang/Throwable'
            self.exceptionhandlers.append((start, end, self.blockd[handler], catchtype))
        self.exceptionhandlers.append((0, 65536, self.rethrowBlock, 'java/lang/Throwable'))

        # State variables for the append/builder loop
        self.current_block = self.entryBlock
        self.current_slots = slotsRvals(self.current_block.inslots)
        for node in self.iNodes:
            # First do a quick check if we have to start a new block
            if not self._canContinueBlock(node):
                self._startNewBlock(node.key)
            vals, outslot_norm = self._getInstrLine(node)

            # Disable exception pruning
            if opts and not vals.jump:
                dummyvals = ResultDict(line=ssa_ops.MagicThrow(self.parent))
                if not self._canAppendInstrToCurrent(node.key, dummyvals):
                    self._startNewBlock(node.key)
                assert self._canAppendInstrToCurrent(node.key, dummyvals)
                self._appendInstr(node, dummyvals, self.current_slots, check_terminate=False)
                vals, outslot_norm = self._getInstrLine(node)

            if not self._canAppendInstrToCurrent(node.key, vals):
                self._startNewBlock(node.key)
                vals, outslot_norm = self._getInstrLine(node)

            assert self._canAppendInstrToCurrent(node.key, vals)
            self._appendInstr(node, vals, outslot_norm)

        # do sanity checks
        assert len(self.blocks) == len(self.blockd)
        for block in self.blocks:
            assert block.jump is not None and block.phis is not None
            assert len(block.predecessors) == len(set(block.predecessors))
            # cleanup temp vars
            block.inslots = None
            block.throwvars = None
            block.chpairs = None
            block.except_used = None
            block.locals_at_except = None

    def _canContinueBlock(self, node):
        return (node.key not in self.blockd) and self.current_block.jump is None # fallthrough goto left as None

    def _chPairsAt(self, address):
        chpairs = []
        for (start, end, handler, catchtype) in self.exceptionhandlers:
            if start <= address < end:
                chpairs.append((catchtype, handler))
        return chpairs

    def _canAppendInstrToCurrent(self, address, vals):
        # If appending exception line to block with existing exceptions, make sure the handlers are the same
        # Also make sure that locals are compatible with all other exceptions in the block
        # If appending a jump, make sure there is no existing exceptions
        block = self.current_block
        if block.chpairs is not None:
            if vals.jump:
                return False
            if vals.line is not None and vals.line.outException is not None:
                chpairs = self._chPairsAt(address)
                if chpairs != block.chpairs:
                    return False

                newlocals = {i: self.current_slots.locals[i] for i in block.except_used}
                return newlocals == block.locals_at_except
        assert block.jump is None
        return True

    def pruneUnused(self, key, newlocals):
        used = toBits(self.used_locals[key])
        return {i: newlocals[i] for i in used}

    def _startNewBlock(self, key):
        ''' We can't continue appending to the current block, so start a new one (or use existing one at location) '''
        # Make new block
        if key not in self.blockd:
            self.makeBlock(key)

        # Finish current block
        block = self.current_block
        curslots = self.current_slots
        assert block.key != key
        if block.jump is None:
            if block.chpairs is not None:
                assert block.throwvars
                self._addOnException(block, self.blockd[key], curslots)
            else:
                assert not block.throwvars
                block.jump = ssa_jumps.Goto(self.parent, self.blockd[key])

        if curslots is not None:
            self.mergeIn((block, False), key, curslots)

        # Update state
        self.current_block = self.blockd[key]
        self.current_slots = slotsRvals(self.current_block.inslots)

    def _getInstrLine(self, iNode):
        parent, initMap = self.parent, self.initMap
        inslots = self.current_slots
        instr = iNode.instruction

        # internal variables won't have any preset type info associated, so we should add in the info from the verifier
        assert len(inslots.stack) == len(iNode.state.stack)
        for i, ivar in enumerate(inslots.stack):
            if ivar and ivar.type == SSA_OBJECT and ivar.decltype is None:
                parent.setObjVarData(ivar, iNode.state.stack[i], initMap)

        for i, ivar in inslots.locals.items():
            if ivar and ivar.type == SSA_OBJECT and ivar.decltype is None:
                parent.setObjVarData(ivar, iNode.state.locals[i], initMap)

        vals = instructionHandlers[instr[0]](self, inslots, iNode)
        newstack = vals.newstack if vals.newstack is not None else inslots.stack
        newlocals = vals.newlocals if vals.newlocals is not None else inslots.locals
        outslot_norm = slots_t(locals=newlocals, stack=newstack)
        return vals, outslot_norm

    def _addOnException(self, block, fallthrough, outslot_norm):
        parent = self.parent
        assert block.throwvars and block.chpairs is not None
        ephi = ssa_ops.ExceptionPhi(parent, block.throwvars)
        block.lines.append(ephi)

        assert block.jump is None
        block.jump = ssa_jumps.OnException(parent, ephi.outException, block.chpairs, fallthrough)
        outslot_except = slots_t(locals=block.locals_at_except, stack=[ephi.outException])
        for suc in block.jump.getExceptSuccessors():
            self.mergeIn((block, True), suc.key, outslot_except)

    def _appendInstr(self, iNode, vals, outslot_norm, check_terminate=True):
        parent = self.parent
        block = self.current_block
        line, jump = vals.line, vals.jump
        if line is not None:
            block.lines.append(line)
        assert block.jump is None
        block.jump = jump

        if line is not None and line.outException is not None:
            block.throwvars.append(line.outException)
            inslots = self.current_slots

            if block.chpairs is None:
                block.chpairs = self._chPairsAt(iNode.key)
                temp = (self.used_locals[h.key] for t, h in block.chpairs)
                block.except_used = toBits(reduce(operator.__or__, temp, 0))
                block.locals_at_except = {i: inslots.locals[i] for i in block.except_used}

            if check_terminate:
                # Return and Throw must be immediately ended because they don't have normal fallthrough
                # CheckCast must terminate block because cast type hack later on requires casts to be at end of block
                if iNode.instruction[0] in (vops.RETURN, vops.THROW) or isinstance(line, ssa_ops.CheckCast):
                    fallthrough = self.getExceptFallthrough(iNode)
                    self._addOnException(block, fallthrough, outslot_norm)

        if block.jump is None:
            unmerged_slots = outslot_norm
        else:
            assert isinstance(block.jump, ssa_jumps.OnException) or not block.throwvars
            unmerged_slots = None
            # Make sure that branch targets are distinct, since this is assumed everywhere
            # Only necessary for if statements as the other jumps merge targets automatically
            # If statements with both branches jumping to same target are replaced with gotos
            block.jump = block.jump.reduceSuccessors([])

            if isinstance(block.jump, subproc.ProcCallOp):
                self.mergeJSROut(iNode, block, outslot_norm)
            else:
                for suc in block.jump.getNormalSuccessors():
                    self.mergeIn((block, False), suc.key, outslot_norm)
        self.current_slots = unmerged_slots
        assert (block.chpairs is None) == (block.except_used is None) == (block.locals_at_except is None)

    def mergeIn(self, from_key, target_key, outslots):
        inslots = self.blockd[target_key].inslots

        assert len(inslots.stack) == len(outslots.stack)
        for i, phi in enumerate(inslots.stack):
            if phi is not None:
                phi.add(from_key, outslots.stack[i])

        for i, phi in inslots.locals.items():
            if phi is not None:
                phi.add(from_key, outslots.locals[i])

        self.blockd[target_key].predecessors.append(from_key)

    ## Block Creation #########################################
    def _makePhiFromVType(self, block, vt):
        var = self.parent.makeVarFromVtype(vt, self.initMap)
        return None if var is None else ssa_ops.Phi(block, var)

    def makeBlockWithInslots(self, key, newlocals, stack):
        assert key not in self.blockd
        block = BasicBlock(key)
        self.blocks.append(block)
        self.blockd[key] = block

        # create inslot phis
        stack = [self._makePhiFromVType(block, vt) for vt in stack]
        newlocals = dict(enumerate(self._makePhiFromVType(block, vt) for vt in newlocals))
        newlocals = self.pruneUnused(key, newlocals)

        block.inslots = slots_t(locals=newlocals, stack=stack)
        block.phis = [phi for phi in stack + block.inslots.localsAsList if phi is not None]
        return block

    def makeBlock(self, key):
        node = self.iNodeD[key]
        return self.makeBlockWithInslots(key, node.state.locals, node.state.stack)

    ###########################################################
    def getExceptFallthrough(self, iNode):
        vop = iNode.instruction[0]
        if vop == vops.RETURN:
            return self.blockd[RETURN_KEY]
        elif vop == vops.THROW:
            return None
        key = iNode.successors[0]
        if key not in self.blockd:
            self.makeBlock(key)
        return self.blockd[key]

    def mergeJSROut(self, jsrnode, block, outslot_norm):
        retnode = self.iNodeD[jsrnode.returnedFrom]
        jump = block.jump
        target_key, ft_key = jump.target.key, jump.fallthrough.key
        assert ft_key == jsrnode.next_instruction

        # first merge regular jump to target
        self.mergeIn((block, False), target_key, outslot_norm)
        # create merged outslots for fallthrough
        fromcall = jump.output
        mask = [mask for key, mask in retnode.state.masks if key == target_key][0]

        skiplocs = fromcall.locals
        retlocs = outslot_norm.locals
        merged = {i: (skiplocs.get(i) if i in mask else retlocs.get(i)) for i in (mask | frozenset(retlocs))}
        # jump.debug_skipvars = set(merged) - set(locals)

        outslot_merged = slots_t(locals=merged, stack=fromcall.stack)
        # merge merged outputs with fallthrough
        self.mergeIn((block, False), ft_key, outslot_merged)



OP_NONE = frozenset(['nop', 'aconst_null', 'iconst_m1', 'iconst_0', 'iconst_1', 'iconst_2', 'iconst_3', 'iconst_4', 'iconst_5', 'lconst_0', 'lconst_1', 'fconst_0', 'fconst_1', 'fconst_2', 'dconst_0', 'dconst_1', 'iload_0', 'iload_1', 'iload_2', 'iload_3', 'lload_0', 'lload_1', 'lload_2', 'lload_3', 'fload_0', 'fload_1', 'fload_2', 'fload_3', 'dload_0', 'dload_1', 'dload_2', 'dload_3', 'aload_0', 'aload_1', 'aload_2', 'aload_3', 'iaload', 'laload', 'faload', 'daload', 'aaload', 'baload', 'caload', 'saload', 'istore_0', 'istore_1', 'istore_2', 'istore_3', 'lstore_0', 'lstore_1', 'lstore_2', 'lstore_3', 'fstore_0', 'fstore_1', 'fstore_2', 'fstore_3', 'dstore_0', 'dstore_1', 'dstore_2', 'dstore_3', 'astore_0', 'astore_1', 'astore_2', 'astore_3', 'iastore', 'lastore', 'fastore', 'dastore', 'aastore', 'bastore', 'castore', 'sastore', 'pop', 'pop2', 'dup', 'dup_x1', 'dup_x2', 'dup2', 'dup2_x1', 'dup2_x2', 'swap', 'iadd', 'ladd', 'fadd', 'dadd', 'isub', 'lsub', 'fsub', 'dsub', 'imul', 'lmul', 'fmul', 'dmul', 'idiv', 'ldiv', 'fdiv', 'ddiv', 'irem', 'lrem', 'frem', 'drem', 'ineg', 'lneg', 'fneg', 'dneg', 'ishl', 'lshl', 'ishr', 'lshr', 'iushr', 'lushr', 'iand', 'land', 'ior', 'lor', 'ixor', 'lxor', 'i2l', 'i2f', 'i2d', 'l2i', 'l2f', 'l2d', 'f2i', 'f2l', 'f2d', 'd2i', 'd2l', 'd2f', 'i2b', 'i2c', 'i2s', 'lcmp', 'fcmpl', 'fcmpg', 'dcmpl', 'dcmpg', 'ireturn', 'lreturn', 'freturn', 'dreturn', 'areturn', 'return', 'arraylength', 'athrow', 'monitorenter', 'monitorexit'])
OP_SHORT = frozenset(['iload', 'lload', 'fload', 'dload', 'aload', 'istore', 'lstore', 'fstore', 'dstore', 'astore', 'ret'])
OP_LBL = frozenset(['ifeq', 'ifne', 'iflt', 'ifge', 'ifgt', 'ifle', 'if_icmpeq', 'if_icmpne', 'if_icmplt', 'if_icmpge', 'if_icmpgt', 'if_icmple', 'if_acmpeq', 'if_acmpne', 'goto', 'jsr', 'ifnull', 'ifnonnull', 'goto_w', 'jsr_w'])
OP_CLS = frozenset(['new', 'anewarray', 'checkcast', 'instanceof'])

OP_FMIM_TO_GUESS = {
    'getstatic': 'Field',
    'putstatic': 'Field',
    'getfield': 'Field',
    'putfield': 'Field',
    'invokevirtual': 'Method',
    'invokespecial': 'Method',
    'invokestatic': 'Method',
}
OP_FMIM = frozenset(OP_FMIM_TO_GUESS)

OPNAMES = ('nop', 'aconst_null', 'iconst_m1', 'iconst_0', 'iconst_1', 'iconst_2', 'iconst_3', 'iconst_4', 'iconst_5', 'lconst_0', 'lconst_1', 'fconst_0', 'fconst_1', 'fconst_2', 'dconst_0', 'dconst_1', 'bipush', 'sipush', 'ldc', 'ldc_w', 'ldc2_w', 'iload', 'lload', 'fload', 'dload', 'aload', 'iload_0', 'iload_1', 'iload_2', 'iload_3', 'lload_0', 'lload_1', 'lload_2', 'lload_3', 'fload_0', 'fload_1', 'fload_2', 'fload_3', 'dload_0', 'dload_1', 'dload_2', 'dload_3', 'aload_0', 'aload_1', 'aload_2', 'aload_3', 'iaload', 'laload', 'faload', 'daload', 'aaload', 'baload', 'caload', 'saload', 'istore', 'lstore', 'fstore', 'dstore', 'astore', 'istore_0', 'istore_1', 'istore_2', 'istore_3', 'lstore_0', 'lstore_1', 'lstore_2', 'lstore_3', 'fstore_0', 'fstore_1', 'fstore_2', 'fstore_3', 'dstore_0', 'dstore_1', 'dstore_2', 'dstore_3', 'astore_0', 'astore_1', 'astore_2', 'astore_3', 'iastore', 'lastore', 'fastore', 'dastore', 'aastore', 'bastore', 'castore', 'sastore', 'pop', 'pop2', 'dup', 'dup_x1', 'dup_x2', 'dup2', 'dup2_x1', 'dup2_x2', 'swap', 'iadd', 'ladd', 'fadd', 'dadd', 'isub', 'lsub', 'fsub', 'dsub', 'imul', 'lmul', 'fmul', 'dmul', 'idiv', 'ldiv', 'fdiv', 'ddiv', 'irem', 'lrem', 'frem', 'drem', 'ineg', 'lneg', 'fneg', 'dneg', 'ishl', 'lshl', 'ishr', 'lshr', 'iushr', 'lushr', 'iand', 'land', 'ior', 'lor', 'ixor', 'lxor', 'iinc', 'i2l', 'i2f', 'i2d', 'l2i', 'l2f', 'l2d', 'f2i', 'f2l', 'f2d', 'd2i', 'd2l', 'd2f', 'i2b', 'i2c', 'i2s', 'lcmp', 'fcmpl', 'fcmpg', 'dcmpl', 'dcmpg', 'ifeq', 'ifne', 'iflt', 'ifge', 'ifgt', 'ifle', 'if_icmpeq', 'if_icmpne', 'if_icmplt', 'if_icmpge', 'if_icmpgt', 'if_icmple', 'if_acmpeq', 'if_acmpne', 'goto', 'jsr', 'ret', 'tableswitch', 'lookupswitch', 'ireturn', 'lreturn', 'freturn', 'dreturn', 'areturn', 'return', 'getstatic', 'putstatic', 'getfield', 'putfield', 'invokevirtual', 'invokespecial','invokestatic', 'invokeinterface', 'invokedynamic', 'new', 'newarray', 'anewarray', 'arraylength', 'athrow', 'checkcast', 'instanceof', 'monitorenter', 'monitorexit', 'wide', 'multianewarray', 'ifnull', 'ifnonnull', 'goto_w', 'jsr_w')
OPNAME_TO_BYTE = {v:i for i, v in enumerate(OPNAMES)}



import collections

from . import bytecode
from .attributes_raw import fixAttributeNames, get_attributes_raw
from .classfileformat.reader import Reader

exceptionHandlerRaw = collections.namedtuple("exceptionHandlerRaw",
                                             ["start","end","handler","type_ind"])

class Code(object):
    def __init__(self, method, bytestream, keepRaw):
        self.method = method
        self.class_ = method.class_

        # Old versions use shorter fields for stack, locals, and code length
        field_fmt = ">HHL" if self.class_.version > (45,2) else ">BBH"
        self.stack, self.locals, codelen = bytestream.get(field_fmt)
        # assert codelen > 0 and codelen < 65536
        self.bytecode_raw = bytestream.getRaw(codelen)
        self.codelen = codelen

        except_cnt = bytestream.get('>H')
        self.except_raw = [bytestream.get('>HHHH') for _ in range(except_cnt)]
        self.except_raw = [exceptionHandlerRaw(*t) for t in self.except_raw]
        attributes_raw = get_attributes_raw(bytestream)
        assert bytestream.size() == 0

        if self.except_raw:
            assert self.stack >= 1

        # print 'Parsing code for', method.name, method.descriptor, method.flags
        codestream = Reader(data=self.bytecode_raw)
        self.bytecode = bytecode.parseInstructions(codestream, self.isIdConstructor)
        self.attributes = fixAttributeNames(attributes_raw, self.class_.cpool)

        for e in self.except_raw:
            assert e.start in self.bytecode
            assert e.end == codelen or e.end in self.bytecode
            assert e.handler in self.bytecode
        if keepRaw:
            self.attributes_raw = attributes_raw

    # This is a callback passed to the bytecode parser to determine if a given method id represents a constructor
    def isIdConstructor(self, methId):
        args = self.class_.cpool.getArgsCheck('Method', methId)
        return args[1] == '<init>'


    def __str__(self):   # pragma: no cover
        lines = ['Stack: {}, Locals {}'.format(self.stack, self.locals)]

        instructions = self.bytecode
        lines += ['{}: {}'.format(i, bytecode.printInstruction(instructions[i])) for i in sorted(instructions)]
        if self.except_raw:
            lines += ['Exception Handlers:']
            lines += map(str, self.except_raw)
        return '\n'.join(lines)

class Method(object):
    flagVals = {'PUBLIC':0x0001,
                'PRIVATE':0x0002,
                'PROTECTED':0x0004,
                'STATIC':0x0008,
                'FINAL':0x0010,
                'SYNCHRONIZED':0x0020,
                'BRIDGE':0x0040,
                'VARARGS':0x0080,
                'NATIVE':0x0100,
                'ABSTRACT':0x0400,
                'STRICTFP':0x0800,
                'SYNTHETIC':0x1000,
                }

    def __init__(self, data, classFile, keepRaw):
        self.class_ = classFile
        cpool = self.class_.cpool

        flags, name_id, desc_id, attributes_raw = data

        self.name = cpool.getArgsCheck('Utf8', name_id)
        self.descriptor = cpool.getArgsCheck('Utf8', desc_id)
        # print 'Loading method ', self.name, self.descriptor
        self.attributes = fixAttributeNames(attributes_raw, cpool)

        self.flags = set(name for name, mask in Method.flagVals.items() if (mask & flags))
        # Flags are ignored for <clinit>?
        if self.name == '<clinit>':
            self.flags = set(['STATIC'])


        self._checkFlags()
        self.static = 'STATIC' in self.flags
        self.native = 'NATIVE' in self.flags
        self.abstract = 'ABSTRACT' in self.flags
        self.isConstructor = (self.name == '<init>')

        self.code = self._loadCode(keepRaw)
        if keepRaw:
            self.attributes_raw = attributes_raw
            self.name_id, self.desc_id = name_id, desc_id

    def _checkFlags(self):
        assert len(self.flags & set(('PRIVATE','PROTECTED','PUBLIC'))) <= 1
        if 'ABSTRACT' in self.flags:
            assert not self.flags & set(['SYNCHRONIZED', 'PRIVATE', 'FINAL', 'STRICT', 'STATIC', 'NATIVE'])

    def _loadCode(self, keepRaw):
        code_attrs = [a for a in self.attributes if a[0] == 'Code']
        if not (self.native or self.abstract):
            assert len(code_attrs) == 1
            code_raw = code_attrs[0][1]
            bytestream = Reader(code_raw)
            return Code(self, bytestream, keepRaw)
        assert not code_attrs
        return None



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2013, Nahuel Riva
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice,this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

__revision__ = "$Id$"

__all__ = ['metadata', 'setup']

from distutils.core import setup
from distutils import version
from warnings import warn

import re
import os
import sys
import glob

# Get the base directory
here = os.path.dirname(__file__)
if not here:
    here = os.path.curdir

# Text describing the module (reStructured text)
try:
    readme = os.path.join(here, 'README.TXT')
    long_description = open(readme, 'r').read()
except Exception:
    warn("README file not found or unreadable!")
    long_description = """Krakatau-noff is a collection of analysis tools for Java .class files."""

# Get the list of scripts
scripts = [os.path.join(here, scr) for scr in ['assemble.py', 'disassemble.py', 'decompile.py']]

# find Krakatau -type d
_packages = """Krakatau
Krakatau/ssa
Krakatau/ssa/constraints
Krakatau/ssa/ssa_jumps
Krakatau/ssa/ssa_ops
Krakatau/verifier
Krakatau/util
Krakatau/classfileformat
Krakatau/assembler
Krakatau/java""".split('\n')

# Set the parameters for the setup script
metadata = {

    # Setup instructions
    'provides'          : ['Krakatau'],
    'packages'          : _packages,
    'scripts'           : scripts,

    # Metadata
    'name'              : 'Krakatau-noff',
    'version'           : 'v0.20181212',
    'description'       : 'Analysis tools for Java .class files.',
    'long_description'  : long_description,
    'url'               : 'https://github.com/drstrng/Krakatau-noff',
    'download_url'      : 'https://github.com/drstrng/Krakatau-noff/archive/v0.20181212.tar.gz',
    'keywords'          : ['java', 'jvm', 'class', '.class', 'assembler', 'disassembler', 'decompiler'],
    }

# Execute the setup script
if __name__ == '__main__':
    setup(**metadata)



from ..verifier import verifier_types as vtypes

# types are represented by classname, dimension
# primitive types are .int, etc since these cannot be valid classnames since periods are forbidden
def TypeTT(baset, dim):
    assert dim >= 0
    return baset, dim

# Not real types
VoidTT = TypeTT('.void', 0)
NullTT = TypeTT('.null', 0)

ObjectTT = TypeTT('java/lang/Object', 0)
StringTT = TypeTT('java/lang/String', 0)
ThrowableTT = TypeTT('java/lang/Throwable', 0)
ClassTT = TypeTT('java/lang/Class', 0)

BoolTT = TypeTT('.boolean', 0)
IntTT = TypeTT('.int', 0)
LongTT = TypeTT('.long', 0)
FloatTT = TypeTT('.float', 0)
DoubleTT = TypeTT('.double', 0)

ByteTT = TypeTT('.byte', 0)
CharTT = TypeTT('.char', 0)
ShortTT = TypeTT('.short', 0)

BExpr = '.bexpr' # bool or byte

def baset(tt): return tt[0]
def dim(tt): return tt[1]
def withDimInc(tt, inc): return TypeTT(baset(tt), dim(tt)+inc)
def withNoDim(tt): return TypeTT(baset(tt), 0)

def isBaseTClass(tt): return not baset(tt).startswith('.')
def className(tt): return baset(tt) if not baset(tt).startswith('.') else None
def primName(tt): return baset(tt)[1:] if baset(tt).startswith('.') else None

###############################################################################

def isSubtype(env, x, y):
    if x == y or y == ObjectTT or x == NullTT:
        return True
    elif y == NullTT:
        return False

    xname, xdim = baset(x), dim(x)
    yname, ydim = baset(y), dim(y)
    if ydim > xdim:
        return False
    elif xdim > ydim: # TODO - these constants should be defined in one place to reduce risk of typos
        return yname in ('java/lang/Object','java/lang/Cloneable','java/io/Serializable')
    else:
        return isBaseTClass(x) and isBaseTClass(y) and env.isSubclass(xname, yname)

# Will not return interface unless all inputs are same interface or null
def commonSupertype(env, tts):
    assert(hasattr(env, 'getClass')) # catch common errors where we forget the env argument

    tts = set(tts)
    tts.discard(NullTT)

    if len(tts) == 1:
        return tts.pop()
    elif not tts:
        return NullTT

    dims = map(dim, tts)
    newdim = min(dims)
    if max(dims) > newdim or any(baset(tt) == 'java/lang/Object' for tt in tts):
        return TypeTT('java/lang/Object', newdim)
    # if any are primitive arrays, result is object array of dim-1
    if not all(isBaseTClass(tt) for tt in tts):
        return TypeTT('java/lang/Object', newdim-1)

    # find common superclass of base types
    bases = sorted(map(baset, tts))
    superclass = reduce(env.commonSuperclass, bases)
    return TypeTT(superclass, newdim)

######################################################################################################
_verifierConvert = {vtypes.T_INT:IntTT, vtypes.T_FLOAT:FloatTT, vtypes.T_LONG:LongTT,
        vtypes.T_DOUBLE:DoubleTT, vtypes.T_SHORT:ShortTT, vtypes.T_CHAR:CharTT,
        vtypes.T_BYTE:ByteTT, vtypes.T_BOOL:BoolTT, vtypes.T_NULL:NullTT,
        vtypes.OBJECT_INFO:ObjectTT}

def verifierToSynthetic_seq(vts):
    return [verifierToSynthetic(vt) for vt in vts if vt != vtypes.T_INVALID]

def verifierToSynthetic(vtype):
    assert vtype.tag not in (None, '.address', '.new', '.init')
    vtype = vtypes.withNoConst(vtype)

    if vtype in _verifierConvert:
        return _verifierConvert[vtype]

    base = vtypes.withNoDimension(vtype)
    if base in _verifierConvert:
        return withDimInc(_verifierConvert[base], vtype.dim)

    return TypeTT(vtype.extra, vtype.dim)

# returns supers, exacts
def declTypeToActual(env, decltype):
    name, newdim = baset(decltype), dim(decltype)

    # Verifier treats bool[]s and byte[]s as interchangeable, so it could really be either
    if newdim and (name == baset(ByteTT) or name == baset(BoolTT)):
        return [], [withDimInc(ByteTT, newdim), withDimInc(BoolTT, newdim)]
    elif not isBaseTClass(decltype): # primitive types can't be subclassed anyway
        return [], [decltype]

    # Verifier doesn't fully verify interfaces so they could be anything
    if env.isInterface(name):
        return [withDimInc(ObjectTT, newdim)], []
    # If class is final, return it as exact, not super
    elif env.isFinal(name):
        return [], [decltype]
    else:
        return [decltype], []

def removeInterface(env, decltype):
    name, newdim = baset(decltype), dim(decltype)
    if isBaseTClass(decltype) and env.isInterface(name):
        return withDimInc(ObjectTT, newdim)
    return decltype



from __future__ import division

from . import opnames

def parseInstructions(bytestream, isConstructor):
    data = bytestream
    assert data.off == 0

    instructions = {}
    while data.size() > 0:
        address = data.off
        inst = getNextInstruction(data, address)

        # replace constructor invocations with synthetic op invokeinit to simplfy things later
        if inst[0] == opnames.INVOKESPECIAL and isConstructor(inst[1]):
            inst = (opnames.INVOKEINIT,) + inst[1:]

        instructions[address] = inst
    assert data.size() == 0
    return instructions

simpleOps = {0x00:opnames.NOP, 0x01:opnames.CONSTNULL, 0x94:opnames.LCMP,
             0xbe:opnames.ARRLEN, 0xbf:opnames.THROW, 0xc2:opnames.MONENTER,
             0xc3:opnames.MONEXIT, 0x57:opnames.POP, 0x58:opnames.POP2, 0x59:opnames.DUP,
             0x5a:opnames.DUPX1, 0x5b:opnames.DUPX2, 0x5c:opnames.DUP2,
             0x5d:opnames.DUP2X1, 0x5e:opnames.DUP2X2, 0x5f:opnames.SWAP}

singleIndexOps = {0xb2:opnames.GETSTATIC,0xb3:opnames.PUTSTATIC,0xb4:opnames.GETFIELD,
            0xb5:opnames.PUTFIELD,0xb6:opnames.INVOKEVIRTUAL,0xb7:opnames.INVOKESPECIAL,
            0xb8:opnames.INVOKESTATIC, 0xbb:opnames.NEW,0xbd:opnames.ANEWARRAY,
            0xc0:opnames.CHECKCAST,0xc1:opnames.INSTANCEOF}

def getNextInstruction(data, address):
    byte = data.get('>B')

    # typecode - B,C,S, and Bool are only used for array types and sign extension
    A,B,C,D,F,I,L,S = "ABCDFIJS"
    Bool = "Z"

    if byte in simpleOps:
        inst = (simpleOps[byte],)
    elif byte in singleIndexOps:
        inst = (singleIndexOps[byte], data.get('>H'))
    elif byte <= 0x11:
        op = opnames.CONST
        if byte <= 0x08:
            t, val = I, byte - 0x03
        elif byte <= 0x0a:
            t, val = L, byte - 0x09
        elif byte <= 0x0d:
            t, val = F, float(byte - 0x0b)
        elif byte <= 0x0f:
            t, val = D, float(byte - 0x0e)
        elif byte == 0x10:
            t, val = I, data.get('>b')
        else:
            t, val = I, data.get('>h')
        inst = op, t, val
    elif byte == 0x12:
        inst = opnames.LDC, data.get('>B'), 1
    elif byte == 0x13:
        inst = opnames.LDC, data.get('>H'), 1
    elif byte == 0x14:
        inst = opnames.LDC, data.get('>H'), 2
    elif byte <= 0x2d:
        op = opnames.LOAD
        if byte <= 0x19:
            t = [I,L,F,D,A][byte - 0x15]
            val = data.get('>B')
        else:
            temp = byte - 0x1a
            t = [I,L,F,D,A][temp // 4]
            val = temp % 4
        inst = op, t, val
    elif byte <= 0x35:
        op = opnames.ARRLOAD
        t = [I,L,F,D,A,B,C,S][byte - 0x2e]
        inst = (op, t) if t != A else (opnames.ARRLOAD_OBJ,) # split object case into seperate op name to simplify things later
    elif byte <= 0x4e:
        op = opnames.STORE
        if byte <= 0x3a:
            t = [I,L,F,D,A][byte - 0x36]
            val = data.get('>B')
        else:
            temp = byte - 0x3b
            t = [I,L,F,D,A][temp // 4]
            val = temp % 4
        inst = op, t, val
    elif byte <= 0x56:
        op = opnames.ARRSTORE
        t = [I,L,F,D,A,B,C,S][byte - 0x4f]
        inst = (op, t) if t != A else (opnames.ARRSTORE_OBJ,) # split object case into seperate op name to simplify things later
    elif byte <= 0x77:
        temp = byte - 0x60
        opt = (opnames.ADD,opnames.SUB,opnames.MUL,opnames.DIV,opnames.REM,opnames.NEG)[temp//4]
        t = (I,L,F,D)[temp % 4]
        inst = opt, t
    elif byte <= 0x83:
        temp = byte - 0x78
        opt = (opnames.SHL,opnames.SHR,opnames.USHR,opnames.AND,opnames.OR,opnames.XOR)[temp//2]
        t = (I,L)[temp % 2]
        inst = opt, t
    elif byte == 0x84:
        inst = opnames.IINC, data.get('>B'), data.get('>b')
    elif byte <= 0x90:
        op = opnames.CONVERT
        pairs = ((I,L),(I,F),(I,D),(L,I),(L,F),(L,D),(F,I),(F,L),(F,D),
                (D,I),(D,L),(D,F))
        src_t, dest_t = pairs[byte - 0x85]
        inst = op, src_t, dest_t
    elif byte <= 0x93:
        op = opnames.TRUNCATE
        dest_t = [B,C,S][byte - 0x91]
        inst = op, dest_t
    elif byte <= 0x98:
        op = opnames.FCMP
        temp = byte - 0x95
        t = (F,D)[temp//2]
        NaN_val = (-1,1)[temp % 2]
        inst = op, t, NaN_val
    elif byte <= 0x9e:
        op = opnames.IF_I
        cmp_t = ('eq','ne','lt','ge','gt','le')[byte - 0x99]
        jumptarget = data.get('>h') + address
        inst = op, cmp_t, jumptarget
    elif byte <= 0xa4:
        op = opnames.IF_ICMP
        cmp_t = ('eq','ne','lt','ge','gt','le')[byte - 0x9f]
        jumptarget = data.get('>h') + address
        inst = op, cmp_t, jumptarget
    elif byte <= 0xa6:
        op = opnames.IF_ACMP
        cmp_t = ('eq','ne')[byte - 0xa5]
        jumptarget = data.get('>h') + address
        inst = op, cmp_t, jumptarget
    elif byte == 0xa7:
        inst = opnames.GOTO, data.get('>h') + address
    elif byte == 0xa8:
        inst = opnames.JSR, data.get('>h') + address
    elif byte == 0xa9:
        inst = opnames.RET, data.get('>B')
    elif byte == 0xaa: # Table Switch
        padding = data.getRaw((3-address) % 4)
        default = data.get('>i') + address
        low = data.get('>i')
        high = data.get('>i')
        assert high >= low
        numpairs = high - low + 1
        offsets = [data.get('>i') + address for _ in range(numpairs)]
        jumps = zip(range(low, high+1), offsets)
        inst = opnames.SWITCH, default, jumps
    elif byte == 0xab: # Lookup Switch
        padding = data.getRaw((3-address) % 4)
        default = data.get('>i') + address
        numpairs = data.get('>i')
        assert numpairs >= 0
        pairs = [data.get('>ii') for _ in range(numpairs)]
        jumps = [(x,(y + address)) for x,y in pairs]
        inst = opnames.SWITCH, default, jumps
    elif byte <= 0xb1:
        op = opnames.RETURN
        t = (I,L,F,D,A,None)[byte - 0xac]
        inst = op, t
    elif byte == 0xb9:
        op = opnames.INVOKEINTERFACE
        index = data.get('>H')
        count, zero = data.get('>B'), data.get('>B')
        inst = op, index, count, zero
    elif byte == 0xba:
        op = opnames.INVOKEDYNAMIC
        index = data.get('>H')
        zero = data.get('>H')
        inst = op, index, zero
    elif byte == 0xbc:
        typecode = data.get('>b')
        types = {4:Bool, 5:C, 6:F, 7:D, 8:B, 9:S, 10:I, 11:L}
        t = types.get(typecode)
        inst = opnames.NEWARRAY, t
    elif byte == 0xc4: # wide
        realbyte = data.get('>B')
        if realbyte >= 0x15 and realbyte < 0x1a:
            t = [I,L,F,D,A][realbyte - 0x15]
            inst = opnames.LOAD, t, data.get('>H')
        elif realbyte >= 0x36 and realbyte < 0x3b:
            t = [I,L,F,D,A][realbyte - 0x36]
            inst = opnames.STORE, t, data.get('>H')
        elif realbyte == 0xa9:
            inst = opnames.RET, data.get('>H')
        elif realbyte == 0x84:
            inst = opnames.IINC, data.get('>H'), data.get('>h')
        else:
            assert 0
    elif byte == 0xc5:
        op = opnames.MULTINEWARRAY
        index = data.get('>H')
        dim = data.get('>B')
        inst = op, index, dim
    elif byte <= 0xc7:
        op = opnames.IF_A
        cmp_t = ('eq','ne')[byte - 0xc6]
        jumptarget = data.get('>h') + address
        inst = op, cmp_t, jumptarget
    elif byte == 0xc8:
        inst = opnames.GOTO, data.get('>i') + address
    elif byte == 0xc9:
        inst = opnames.JSR, data.get('>i') + address
    else:
        assert 0
    return inst

def printInstruction(instr):
    if len(instr) == 1:
        return instr[0]
    elif len(instr) == 2:
        return '{}({})'.format(*instr)
    else:
        return '{}{}'.format(instr[0], instr[1:])



from .base import BaseOp

from .array import ArrLoad, ArrStore, ArrLength
from .checkcast import CheckCast, InstanceOf
from .convert import Convert
from .fieldaccess import FieldAccess
from .fmath import FAdd, FDiv, FMul, FRem, FSub, FNeg, FCmp
from .invoke import Invoke, InvokeDynamic
from .imath import IAdd, IDiv, IMul, IRem, ISub, IAnd, IOr, IShl, IShr, IUshr, IXor, ICmp
from .monitor import Monitor
from .new import New, NewArray, MultiNewArray
from .throw import Throw, MagicThrow
from .truncate import Truncate
from .tryreturn import TryReturn

from .phi import Phi, ExceptionPhi



import collections

from .reader import Reader

TAGS = [None, 'Utf8', None, 'Int', 'Float', 'Long', 'Double', 'Class', 'String', 'Field', 'Method', 'InterfaceMethod', 'NameAndType', None, None, 'MethodHandle', 'MethodType', None, 'InvokeDynamic', 'Module', 'Package']

SlotData = collections.namedtuple('SlotData', ['tag', 'data', 'refs'])
ExceptData = collections.namedtuple('ExceptData', ['start', 'end', 'handler', 'type'])

class ConstantPoolData(object):
    def __init__(self, r):
        self.slots = []
        self._null()

        size = r.u16()
        while len(self.slots) < size:
            self._const(r)

    def _null(self):
        self.slots.append(SlotData(None, None, None))

    def _const(self, r):
        t = TAGS[r.u8()]
        data = None
        refs = []

        if t == 'Utf8':
            data = r.getRaw(r.u16())
        elif t == 'Int' or t == 'Float':
            data = r.u32()
        elif t == 'Long' or t == 'Double':
            data = r.u64()
        elif t == 'MethodHandle':
            data = r.u8()
            refs.append(r.u16())
        elif t in ['Class', 'String', 'MethodType', 'Module', 'Package']:
            refs.append(r.u16())
        else:
            refs.append(r.u16())
            refs.append(r.u16())
        self.slots.append(SlotData(t, data, refs))
        if t in ('Long', 'Double'):
            self._null()

    def getutf(self, ind):
        if ind < len(self.slots) and self.slots[ind].tag == 'Utf8':
            return self.slots[ind].data

    def getclsutf(self, ind):
        if ind < len(self.slots) and self.slots[ind].tag == 'Class':
            return self.getutf(self.slots[ind].refs[0])

class BootstrapMethodsData(object):
    def __init__(self, r):
        self.slots = []
        for _ in range(r.u16()):
            first = r.u16()
            argcount = r.u16()
            refs = [first] + [r.u16() for _ in range(argcount)]
            self.slots.append(SlotData('Bootstrap', None, refs))

class CodeData(object):
    def __init__(self, r, pool, short):
        if short:
            self.stack, self.locals, codelen = r.u8(), r.u8(), r.u16()
        else:
            self.stack, self.locals, codelen = r.u16(), r.u16(), r.u32()

        self.bytecode = r.getRaw(codelen)
        self.exceptions = [ExceptData(r.u16(), r.u16(), r.u16(), r.u16()) for _ in range(r.u16())]
        self.attributes = [AttributeData(r) for _ in range(r.u16())]

class AttributeData(object):
    def __init__(self, r, pool=None):
        self.name, self.length = r.u16(), r.u32()

        # The JVM allows InnerClasses attributes to have a bogus length field,
        # and hence we must calculate the length from the contents
        if pool and pool.getutf(self.name) == b'InnerClasses':
            actual_length = r.copy().u16() * 8 + 2
        else:
            actual_length = self.length

        self.raw = r.getRaw(actual_length)
        self.wronglength = actual_length != self.length

    def stream(self): return Reader(self.raw)

class FieldData(object):
    def __init__(self, r):
        self.access, self.name, self.desc = r.u16(), r.u16(), r.u16()
        self.attributes = [AttributeData(r) for _ in range(r.u16())]

class MethodData(object):
    def __init__(self, r):
        self.access, self.name, self.desc = r.u16(), r.u16(), r.u16()
        self.attributes = [AttributeData(r) for _ in range(r.u16())]

class ClassData(object):
    def __init__(self, r):
        magic, minor, major = r.u32(), r.u16(), r.u16()
        self.version = major, minor

        self.pool = ConstantPoolData(r)

        self.access, self.this, self.super = r.u16(), r.u16(), r.u16()
        self.interfaces = [r.u16() for _ in range(r.u16())]
        self.fields = [FieldData(r) for _ in range(r.u16())]
        self.methods = [MethodData(r) for _ in range(r.u16())]
        self.attributes = [AttributeData(r, pool=self.pool) for _ in range(r.u16())]
        # assert r.done()

    def getattrs(self, name):
        for attr in self.attributes:
            if self.pool.getutf(attr.name) == name:
                yield attr



from .. import excepttypes, objtypes
from ..constraints import FloatConstraint, IntConstraint, ObjectConstraint, maybeThrow, returnOrThrow, throw
from ..ssa_types import SSA_INT

from .base import BaseOp

def getElementTypes(env, tops):
    types = [objtypes.withDimInc(tt, -1) for tt in tops]
    # temporary hack
    types = [objtypes.removeInterface(env, tt) for tt in types]

    supers = [tt for tt in types if objtypes.isBaseTClass(tt)]
    exact = [tt for tt in types if not objtypes.isBaseTClass(tt)]
    return ObjectConstraint.fromTops(env, supers, exact)

class ArrLoad(BaseOp):
    def __init__(self, parent, args, ssatype):
        super(ArrLoad, self).__init__(parent, args, makeException=True)
        self.env = parent.env
        self.rval = parent.makeVariable(ssatype, origin=self)
        self.ssatype = ssatype

    def propagateConstraints(self, a, i):
        etypes = (excepttypes.ArrayOOB,)
        if a.null:
            etypes += (excepttypes.NullPtr,)
            if a.isConstNull():
                return throw(ObjectConstraint.fromTops(self.env, [], [excepttypes.NullPtr], nonnull=True))

        if self.ssatype[0] == 'int':
            rout = IntConstraint.bot(self.ssatype[1])
        elif self.ssatype[0] == 'float':
            rout = FloatConstraint.bot(self.ssatype[1])
        elif self.ssatype[0] == 'obj':
            rout = getElementTypes(self.env, a.types.supers | a.types.exact)

        eout = ObjectConstraint.fromTops(self.env, [], etypes, nonnull=True)
        return returnOrThrow(rout, eout)

class ArrStore(BaseOp):
    has_side_effects = True

    def __init__(self, parent, args):
        super(ArrStore, self).__init__(parent, args, makeException=True)
        self.env = parent.env

    def propagateConstraints(self, a, i, x):
        etypes = (excepttypes.ArrayOOB,)
        if a.null:
            etypes += (excepttypes.NullPtr,)
            if a.isConstNull():
                return throw(ObjectConstraint.fromTops(self.env, [], [excepttypes.NullPtr], nonnull=True))

        if isinstance(x, ObjectConstraint):
            # If the type of a is known exactly to be the single possibility T[]
            # and x is assignable to T, we can assume there is no ArrayStore exception
            # if a's type has multiple possibilities, then there can be an exception
            known_type = a.types.exact if len(a.types.exact) == 1 else frozenset()
            allowed = getElementTypes(self.env, known_type)
            if allowed.meet(x) != allowed:
                etypes += (excepttypes.ArrayStore,)

        return maybeThrow(ObjectConstraint.fromTops(self.env, [], etypes, nonnull=True))

class ArrLength(BaseOp):
    def __init__(self, parent, args):
        super(ArrLength, self).__init__(parent, args, makeException=True)
        self.env = parent.env
        self.rval = parent.makeVariable(SSA_INT, origin=self)

    def propagateConstraints(self, x):
        etypes = ()
        if x.null:
            etypes += (excepttypes.NullPtr,)
            if x.isConstNull():
                return throw(ObjectConstraint.fromTops(self.env, [], [excepttypes.NullPtr], nonnull=True))

        excons = ObjectConstraint.fromTops(self.env, [], etypes, nonnull=True)
        return returnOrThrow(IntConstraint.range(32, 0, (1<<31)-1), excons)



from . import objtypes

# common exception types
Arithmetic = objtypes.TypeTT('java/lang/ArithmeticException', 0)
ArrayOOB = objtypes.TypeTT('java/lang/ArrayIndexOutOfBoundsException', 0)
ArrayStore = objtypes.TypeTT('java/lang/ArrayStoreException', 0)
ClassCast = objtypes.TypeTT('java/lang/ClassCastException', 0)
MonState = objtypes.TypeTT('java/lang/IllegalMonitorStateException', 0)
NegArrSize = objtypes.TypeTT('java/lang/NegativeArraySizeException', 0)
NullPtr = objtypes.TypeTT('java/lang/NullPointerException', 0)
OOM = objtypes.TypeTT('java/lang/OutOfMemoryError', 0)



import copy

from ..functionbase import SSAFunctionBase

class BaseJump(SSAFunctionBase):
    def __init__(self, parent, arguments=()):
        super(BaseJump, self).__init__(parent,arguments)

    def replaceBlocks(self, blockDict):
        assert not self.getSuccessors()

    def getNormalSuccessors(self): return []
    def getExceptSuccessors(self): return []
    def getSuccessors(self): return self.getNormalSuccessors() + self.getExceptSuccessors()
    def getSuccessorPairs(self): return [(x,False) for x in self.getNormalSuccessors()] + [(x,True) for x in self.getExceptSuccessors()]
    def reduceSuccessors(self, pairsToRemove): return self

    def clone(self): return copy.copy(self) # overriden by classes which need to do a deep copy



import collections
from functools import partial
import operator

from .. import graph_util
from ..namegen import NameGen
from ..ssa import objtypes
from ..verifier.descriptors import parseMethodDescriptor

from . import ast, ast2, astgen, boolize, graphproxy, mergevariables, structuring

class DeclInfo(object):
    __slots__ = "declScope scope defs".split()
    def __init__(self):
        self.declScope = self.scope = None
        self.defs = []

def findVarDeclInfo(root, predeclared):
    info = collections.OrderedDict()
    def visit(scope, expr):
        for param in expr.params:
            visit(scope, param)

        if expr.isLocalAssign():
            left, right = expr.params
            info[left].defs.append(right)
        elif isinstance(expr, (ast.Local, ast.Literal)):
            # this would be so much nicer if we had Ordered defaultdicts
            info.setdefault(expr, DeclInfo())
            info[expr].scope = ast.StatementBlock.join(info[expr].scope, scope)

    def visitDeclExpr(scope, expr):
        info.setdefault(expr, DeclInfo())
        assert scope is not None and info[expr].declScope is None
        info[expr].declScope = scope

    for expr in predeclared:
        visitDeclExpr(root, expr)

    stack = [(root,root)]
    while stack:
        scope, stmt = stack.pop()
        if isinstance(stmt, ast.StatementBlock):
            stack.extend((stmt,sub) for sub in stmt.statements)
        else:
            stack.extend((subscope,subscope) for subscope in stmt.getScopes())
            # temp hack
            if stmt.expr is not None:
                visit(scope, stmt.expr)
            if isinstance(stmt, ast.TryStatement):
                for catchdecl, body in stmt.pairs:
                    visitDeclExpr(body, catchdecl.local)
    return info

def reverseBoolExpr(expr):
    assert expr.dtype == objtypes.BoolTT
    if isinstance(expr, ast.BinaryInfix):
        symbols = "== != < >= > <=".split()
        floatts = (objtypes.FloatTT, objtypes.DoubleTT)
        if expr.opstr in symbols:
            sym2 = symbols[symbols.index(expr.opstr) ^ 1]
            left, right = expr.params
            # be sure not to reverse floating point comparisons since it's not equivalent for NaN
            if expr.opstr in symbols[:2] or (left.dtype not in floatts and right.dtype not in floatts):
                return ast.BinaryInfix(sym2, [left, right], objtypes.BoolTT)
    elif isinstance(expr, ast.UnaryPrefix) and expr.opstr == '!':
        return expr.params[0]
    return ast.UnaryPrefix('!', expr)

def getSubscopeIter(root):
    stack = [root]
    while stack:
        scope = stack.pop()
        if isinstance(scope, ast.StatementBlock):
            stack.extend(scope.statements)
            yield scope
        else:
            stack.extend(scope.getScopes())

def mayBreakTo(root, forbidden):
    assert None not in forbidden
    for scope in getSubscopeIter(root):
        if scope.jumpKey in forbidden:
            # We return true if scope has forbidden jump and is reachable
            # We assume there is no unreachable code, so in order for a scope
            # jump to be unreachable, it must end in a return, throw, or a
            # compound statement, all of which are not reachable or do not
            # break out of the statement. We omit adding last.breakKey to
            # forbidden since it should always match scope.jumpKey anyway
            if not scope.statements:
                return True
            last = scope.statements[-1]
            if not last.getScopes():
                if not isinstance(last, (ast.ReturnStatement, ast.ThrowStatement)):
                    return True
            else:
                # If and switch statements may allow fallthrough
                # A while statement with condition may break implicitly
                if isinstance(last, ast.IfStatement) and len(last.getScopes()) == 1:
                    return True
                if isinstance(last, ast.SwitchStatement) and not last.hasDefault():
                    return True
                if isinstance(last, ast.WhileStatement) and last.expr != ast.Literal.TRUE:
                    return True

                if not isinstance(last, ast.WhileStatement):
                    for sub in last.getScopes():
                        assert sub.breakKey == last.breakKey == scope.jumpKey
    return False

def replaceKeys(top, replace):
    assert None not in replace
    get = lambda k:replace.get(k,k)

    if top.getScopes():
        if isinstance(top, ast.StatementBlock) and get(top.breakKey) is None:
            # breakKey can be None with non-None jumpKey when we're a scope in a switch statement that falls through
            # and the end of the switch statement is unreachable
            assert get(top.jumpKey) is None or not top.labelable

        top.breakKey = get(top.breakKey)
        if isinstance(top, ast.StatementBlock):
            top.jumpKey = get(top.jumpKey)
            for item in top.statements:
                replaceKeys(item, replace)
        else:
            for scope in top.getScopes():
                replaceKeys(scope, replace)

NONE_SET = frozenset([None])
def _preorder(scope, func):
    newitems = []
    for i, item in enumerate(scope.statements):
        for sub in item.getScopes():
            _preorder(sub, func)

        val = func(scope, item)
        vals = [item] if val is None else val
        newitems.extend(vals)
    scope.statements = newitems

def _fixObjectCreations(scope, item):
    '''Combines new/invokeinit pairs into Java constructor calls'''
    # Thanks to the uninitialized variable merging prior to AST generation,
    # we can safely assume there are no copies to worry about
    expr = item.expr
    if isinstance(expr, ast.Assignment):
        left, right = expr.params
        if isinstance(right, ast.Dummy) and right.isNew:
            return [] # remove item
    elif isinstance(expr, ast.MethodInvocation) and expr.name == '<init>':
        left = expr.params[0]
        newexpr = ast.ClassInstanceCreation(ast.TypeName(left.dtype), expr.tts[1:], expr.params[1:])
        item.expr = ast.Assignment(left, newexpr)

def _pruneRethrow_cb(item):
    '''Convert try{A} catch(T t) {throw t;} to {A}'''
    while item.pairs:
        decl, body = item.pairs[-1]
        caught, lines = decl.local, body.statements

        if len(lines) == 1:
            line = lines[0]
            if isinstance(line, ast.ThrowStatement) and line.expr == caught:
                item.pairs = item.pairs[:-1]
                continue
        break
    if not item.pairs:
        new = item.tryb
        assert new.breakKey == item.breakKey
        assert new.continueKey == item.continueKey
        assert not new.labelable
        new.labelable = True
        return new
    return item

def _pruneIfElse_cb(item):
    '''Convert if(A) {B} else {} to if(A) {B}'''
    if len(item.scopes) > 1:
        tblock, fblock = item.scopes

        # if true block is empty, swap it with false so we can remove it
        if not tblock.statements and tblock.doesFallthrough():
            item.expr = reverseBoolExpr(item.expr)
            tblock, fblock = item.scopes = fblock, tblock

        if not fblock.statements and fblock.doesFallthrough():
            item.scopes = tblock,

    # if(A) {B throw/return ... } else {C} -> if(A) {B throw/return ...} {C}
    if len(item.scopes) > 1:
        # How much we want the block to be first, or (3, 0) if it can't be simplified
        def priority(block):
            # If an empty block survives to this point, it must end in a break so we can simplify in this case too
            if not block.statements:
                return 2, 0
            # If any complex statements, there might be a labeled break, so it's not safe
            if any(stmt.getScopes() for stmt in block.statements):
                return 3, 0
            # prefer if(...) {throw...} return... to if(...) {return...} throw...
            if isinstance(block.statements[-1], ast.ThrowStatement):
                return 0, len(block.statements)
            elif isinstance(block.statements[-1], ast.ReturnStatement):
                return 1, len(block.statements)
            return 3, 0

        if priority(fblock) < priority(tblock):
            item.expr = reverseBoolExpr(item.expr)
            tblock, fblock = item.scopes = fblock, tblock

        if priority(tblock) < (3, 0):
            assert tblock.statements or not tblock.doesFallthrough()
            item.scopes = tblock,
            item.breakKey = fblock.continueKey
            fblock.labelable = True
            return [item], fblock

        # If cond is !(x), reverse it back to simplify cond
        if isinstance(item.expr, ast.UnaryPrefix) and item.expr.opstr == '!':
            item.expr = reverseBoolExpr(item.expr)
            tblock, fblock = item.scopes = fblock, tblock

    # if(A) {if(B) {C}} -> if(A && B) {C}
    tblock = item.scopes[0]
    if len(item.scopes) == 1 and len(tblock.statements) == 1 and tblock.doesFallthrough():
        first = tblock.statements[0]
        if isinstance(first, ast.IfStatement) and len(first.scopes) == 1:
            item.expr = ast.BinaryInfix('&&',[item.expr, first.expr], objtypes.BoolTT)
            item.scopes = first.scopes
    return [], item

def _whileCondition_cb(item):
    '''Convert while(true) {if(A) {B break;} else {C} D} to while(!A) {{C} D} {B}
        and while(A) {if(B) {break;} else {C} D} to while(A && !B) {{C} D}'''
    failure = [], item # what to return if we didn't inline
    body = item.getScopes()[0]
    if not body.statements or not isinstance(body.statements[0], ast.IfStatement):
        return failure

    head = body.statements[0]
    cond = head.expr
    trueb, falseb = (head.getScopes() + (None,))[:2]

    # Make sure it doesn't continue the loop or break out of the if statement
    badjumps1 = frozenset([head.breakKey, item.continueKey]) - NONE_SET
    if mayBreakTo(trueb, badjumps1):
        if falseb is not None and not mayBreakTo(falseb, badjumps1):
            cond = reverseBoolExpr(cond)
            trueb, falseb = falseb, trueb
        else:
            return failure
    assert not mayBreakTo(trueb, badjumps1)


    trivial = not trueb.statements and trueb.jumpKey == item.breakKey
    # If we already have a condition, only a simple break is allowed
    if not trivial and item.expr != ast.Literal.TRUE:
        return failure

    # If break body is nontrival, we can't insert this after the end of the loop unless
    # We're sure that nothing else in the loop breaks out
    badjumps2 = frozenset([item.breakKey]) - NONE_SET
    if not trivial:
        restloop = [falseb] if falseb is not None else []
        restloop += body.statements[1:]
        if body.jumpKey == item.breakKey or any(mayBreakTo(s, badjumps2) for s in restloop):
            return failure

    # Now inline everything
    item.expr = _simplifyExpressions(ast.BinaryInfix('&&', [item.expr, reverseBoolExpr(cond)]))
    if falseb is None:
        body.continueKey = body.statements.pop(0).breakKey
    else:
        body.continueKey = falseb.continueKey
        body.statements[0] = falseb
        falseb.labelable = True
    trueb.labelable = True

    if item.breakKey is None: # Make sure to maintain invariant that bkey=None -> jkey=None
        assert trueb.doesFallthrough()
        trueb.jumpKey = trueb.breakKey = None
    trueb.breakKey = item.breakKey
    assert trueb.continueKey is not None
    if not trivial:
        item.breakKey = trueb.continueKey

    # Trueb doesn't break to head.bkey but there might be unreacahble jumps, so we replace
    # it too. We don't replace item.ckey because it should never appear, even as an
    # unreachable jump
    replaceKeys(trueb, {head.breakKey:trueb.breakKey, item.breakKey:trueb.breakKey})
    return [item], trueb

def _simplifyBlocksSub(scope, item, isLast):
    rest = []
    if isinstance(item, ast.TryStatement):
        item = _pruneRethrow_cb(item)
    elif isinstance(item, ast.IfStatement):
        rest, item = _pruneIfElse_cb(item)
    elif isinstance(item, ast.WhileStatement):
        rest, item = _whileCondition_cb(item)

    if isinstance(item, ast.StatementBlock):
        assert item.breakKey is not None or item.jumpKey is None
        # If bkey is None, it can't be broken to
        # If contents can also break to enclosing scope, it's always safe to inline
        bkey = item.breakKey
        if bkey is None or (bkey == scope.breakKey and scope.labelable):
            rest, item.statements = rest + item.statements, []
        # Now inline statements at the beginning of the scope that don't need to break to it
        for sub in item.statements[:]:
            if sub.getScopes() and sub.breakKey != bkey and mayBreakTo(sub, frozenset([bkey])):
                break
            rest.append(item.statements.pop(0))

        if not item.statements:
            if item.jumpKey != bkey:
                assert isLast
                scope.jumpKey = item.jumpKey
                assert scope.breakKey is not None or scope.jumpKey is None
            return rest
    return rest + [item]

def _simplifyBlocks(scope):
    newitems = []
    for item in reversed(scope.statements):
        isLast = not newitems # may be true if all subsequent items pruned
        if isLast and item.getScopes():
            if item.breakKey != scope.jumpKey:# and item.breakKey is not None:
                replaceKeys(item, {item.breakKey: scope.jumpKey})

        for sub in reversed(item.getScopes()):
            _simplifyBlocks(sub)
        vals = _simplifyBlocksSub(scope, item, isLast)
        newitems += reversed(vals)
    scope.statements = newitems[::-1]


_op2bits = {'==':2, '!=':13, '<':1, '<=':3, '>':4, '>=':6}
_bit2ops_float = {v:k for k,v in _op2bits.items()}
_bit2ops = {(v & 7):k for k,v in _op2bits.items()}

def _getBitfield(expr):
    if isinstance(expr, ast.BinaryInfix):
        if expr.opstr in ('==','!=','<','<=','>','>='):
            # We don't want to merge expressions if they could have side effects
            # so only allow literals and locals
            if all(isinstance(p, (ast.Literal, ast.Local)) for p in expr.params):
                return _op2bits[expr.opstr], list(expr.params)
        elif expr.opstr in ('&','&&','|','||'):
            bits1, args1 = _getBitfield(expr.params[0])
            bits2, args2 = _getBitfield(expr.params[1])
            if args1 == args2:
                bits = (bits1 & bits2) if '&' in expr.opstr else (bits1 | bits2)
                return bits, args1
    elif isinstance(expr, ast.UnaryPrefix) and expr.opstr == '!':
        bits, args = _getBitfield(expr.params[0])
        return ~bits, args
    return 0, None

def _mergeComparisons(expr):
    # a <= b && a != b -> a < b, etc.
    bits, args = _getBitfield(expr)
    if args is None:
        return expr

    assert not hasSideEffects(args[0]) and not hasSideEffects(args[1])
    if args[0].dtype in (objtypes.FloatTT, objtypes.DoubleTT):
        mask, d = 15, _bit2ops_float
    else:
        mask, d = 7, _bit2ops

    bits &= mask
    notbits = (~bits) & mask

    if bits == 0:
        return ast.Literal.TRUE
    elif notbits == 0:
        return ast.Literal.FALSE
    elif bits in d:
        return ast.BinaryInfix(d[bits], args, objtypes.BoolTT)
    elif notbits in d:
        return ast.UnaryPrefix('!', ast.BinaryInfix(d[notbits], args, objtypes.BoolTT))
    return expr

def _simplifyExpressions(expr):
    TRUE, FALSE = ast.Literal.TRUE, ast.Literal.FALSE
    bools = {True:TRUE, False:FALSE}
    opfuncs = {'<': operator.lt, '<=': operator.le, '>': operator.gt, '>=': operator.ge}

    simplify = _simplifyExpressions
    expr.params = map(simplify, expr.params)

    if isinstance(expr, ast.BinaryInfix):
        left, right = expr.params
        op = expr.opstr
        if op in ('==','!=','<','<=','>','>=') and isinstance(right, ast.Literal):
            # la cmp lb -> result (i.e. constant propagation on literal comparisons)
            if isinstance(left, ast.Literal):
                if op in ('==','!='):
                    # these could be string or class literals, but those are always nonnull so it still works
                    res = (left == right) == (op == '==')
                else:
                    assert left.dtype == right.dtype
                    res = opfuncs[op](left.val, right.val)
                expr = bools[res]
            # (a ? lb : c) cmp ld -> a ? (lb cmp ld) : (c cmp ld)
            elif isinstance(left, ast.Ternary) and isinstance(left.params[1], ast.Literal):
                left.params[1] = simplify(ast.BinaryInfix(op, [left.params[1], right], expr._dtype))
                left.params[2] = simplify(ast.BinaryInfix(op, [left.params[2], right], expr._dtype))
                expr = left

    # a ? true : b -> a || b
    # a ? false : b -> !a && b
    if isinstance(expr, ast.Ternary) and expr.dtype == objtypes.BoolTT:
        cond, val1, val2 = expr.params
        if not isinstance(val1, ast.Literal): # try to get bool literal to the front
            cond, val1, val2 = reverseBoolExpr(cond), val2, val1

        if val1 == TRUE:
            expr = ast.BinaryInfix('||', [cond, val2], objtypes.BoolTT)
        elif val1 == FALSE:
            expr = ast.BinaryInfix('&&', [reverseBoolExpr(cond), val2], objtypes.BoolTT)

    # true && a -> a, etc.
    if isinstance(expr, ast.BinaryInfix) and expr.opstr in ('&&','||'):
        left, right = expr.params
        if expr.opstr == '&&':
            if left == TRUE or (right == FALSE and not hasSideEffects(left)):
                expr = right
            elif left == FALSE or right == TRUE:
                expr = left
        else:
            if left == TRUE or right == FALSE:
                expr = left
            elif left == FALSE or (right == TRUE and not hasSideEffects(left)):
                expr = right
        # a > b || a == b -> a >= b, etc.
        expr = _mergeComparisons(expr)

    # a == true -> a
    # a == false -> !a
    if isinstance(expr, ast.BinaryInfix) and expr.opstr in ('==, !=') and expr.params[0].dtype == objtypes.BoolTT:
        left, right = expr.params
        if not isinstance(left, ast.Literal): # try to get bool literal to the front
            left, right = right, left
        if isinstance(left, ast.Literal):
            flip = (left == TRUE) != (expr.opstr == '==')
            expr = reverseBoolExpr(right) if flip else right

    # !a ? b : c -> a ? c : b
    if isinstance(expr, ast.Ternary) and isinstance(expr.params[0], ast.UnaryPrefix):
        cond, val1, val2 = expr.params
        if cond.opstr == '!':
            expr.params = [reverseBoolExpr(cond), val2, val1]

    # 0 - a -> -a
    if isinstance(expr, ast.BinaryInfix) and expr.opstr == '-':
        if expr.params[0] == ast.Literal.ZERO or expr.params[0] == ast.Literal.LZERO:
            expr = ast.UnaryPrefix('-', expr.params[1])

    # (double)4.2f -> 4.2, etc.
    if isinstance(expr, ast.Cast) and isinstance(expr.params[1], ast.Literal):
        expr = ast.makeCastExpr(expr.dtype, expr.params[1])

    return expr

def _setScopeParents(scope):
    for item in scope.statements:
        for sub in item.getScopes():
            sub.bases = scope.bases + (sub,)
            _setScopeParents(sub)

def _replaceExpressions(scope, item, rdict):
    # Must be done before local declarations are created since it doesn't touch/remove them
    if item.expr is not None:
        item.expr = item.expr.replaceSubExprs(rdict)
    # remove redundant assignments i.e. x=x;
    if isinstance(item.expr, ast.Assignment):
        assert isinstance(item, ast.ExpressionStatement)
        left, right = item.expr.params
        if left == right:
            return []
    return [item]

def _oldMergeVariables(root, predeclared):
    _setScopeParents(root)
    info = findVarDeclInfo(root, predeclared)

    lvars = [expr for expr in info if isinstance(expr, ast.Local)]
    forbidden = set()
    # If var has any defs which aren't a literal or local, mark it as a leaf node (it can't be merged into something)
    for var in lvars:
        if not all(isinstance(expr, (ast.Local, ast.Literal)) for expr in info[var].defs):
            forbidden.add(var)
        elif info[var].declScope is not None:
            forbidden.add(var)

    sccs = graph_util.tarjanSCC(lvars, lambda var:([] if var in forbidden else info[var].defs))
    # the sccs will be in topolgical order
    varmap = {}
    for scc in sccs:
        if forbidden.isdisjoint(scc):
            alldefs = []
            for expr in scc:
                for def_ in info[expr].defs:
                    if def_ not in scc:
                        alldefs.append(varmap[def_])
            if len(set(alldefs)) == 1:
                target = alldefs[0]
                if all(var.dtype == target.dtype for var in scc):
                    scope = ast.StatementBlock.join(*(info[var].scope for var in scc))
                    scope = ast.StatementBlock.join(scope, info[target].declScope) # scope is unchanged if declScope is none like usual
                    if info[target].declScope is None or info[target].declScope == scope:
                        for var in scc:
                            varmap[var] = target
                        info[target].scope = ast.StatementBlock.join(scope, info[target].scope)
                        continue
        # fallthrough if merging is impossible
        for var in scc:
            varmap[var] = var
            if len(info[var].defs) > 1:
                forbidden.add(var)
    _preorder(root, partial(_replaceExpressions, rdict=varmap))

def _mergeVariables(root, predeclared, isstatic):
    _oldMergeVariables(root, predeclared)

    rdict = mergevariables.mergeVariables(root, isstatic, predeclared)
    _preorder(root, partial(_replaceExpressions, rdict=rdict))

_oktypes = ast.BinaryInfix, ast.Local, ast.Literal, ast.Parenthesis, ast.Ternary, ast.TypeName, ast.UnaryPrefix
def hasSideEffects(expr):
    if not isinstance(expr, _oktypes):
        return True
    # check for division by 0. If it's a float or dividing by nonzero literal, it's ok
    elif isinstance(expr, ast.BinaryInfix) and expr.opstr in ('/','%'):
        if expr.dtype not in (objtypes.FloatTT, objtypes.DoubleTT):
            divisor = expr.params[-1]
            if not isinstance(divisor, ast.Literal) or divisor.val == 0:
                return True
    return False

def _inlineVariables(root):
    # first find all variables with a single def and use
    defs = collections.defaultdict(list)
    uses = collections.defaultdict(int)

    def visitExprFindDefs(expr):
        if expr.isLocalAssign():
            defs[expr.params[0]].append(expr)
        elif isinstance(expr, ast.Local):
            uses[expr] += 1

    def visitFindDefs(scope, item):
        if item.expr is not None:
            stack = [item.expr]
            while stack:
                expr = stack.pop()
                visitExprFindDefs(expr)
                stack.extend(expr.params)

    _preorder(root, visitFindDefs)
    # These should have 2 uses since the initial assignment also counts
    replacevars = {k for k,v in defs.items() if len(v)==1 and uses[k]==2 and k.dtype == v[0].params[1].dtype}
    def doReplacement(item, pairs):
        old, new = item.expr.params
        assert isinstance(old, ast.Local) and old.dtype == new.dtype
        stack = [(True, (True, item2, expr)) for item2, expr in reversed(pairs) if expr is not None]
        while stack:
            recurse, args = stack.pop()

            if recurse:
                canReplace, parent, expr = args
                stack.append((False, expr))

                # TODO - fix this for real
                if expr.complexity() > 30:
                    canReplace = False

                # For ternaries, we don't want to replace into the conditionally
                # evaluated part, but we still need to check those parts for
                # barriers. For both ternaries and short circuit operators, the
                # first param is always evaluated, so it is safe
                if isinstance(expr, ast.Ternary) or isinstance(expr, ast.BinaryInfix) and expr.opstr in ('&&','||'):
                    for param in reversed(expr.params[1:]):
                        stack.append((True, (False, expr, param)))
                    stack.append((True, (canReplace, expr, expr.params[0])))
                # For assignments, we unroll the LHS arguments, because if assigning
                # to an array or field, we don't want that to serve as a barrier
                elif isinstance(expr, ast.Assignment):
                    left, right = expr.params
                    stack.append((True, (canReplace, expr, right)))
                    if isinstance(left, (ast.ArrayAccess, ast.FieldAccess)):
                        for param in reversed(left.params):
                            stack.append((True, (canReplace, left, param)))
                    else:
                        assert isinstance(left, ast.Local)
                else:
                    for param in reversed(expr.params):
                        stack.append((True, (canReplace, expr, param)))

                if expr == old:
                    if canReplace:
                        if isinstance(parent, ast.JavaExpression):
                            params = parent.params = list(parent.params)
                            params[params.index(old)] = new
                        else: # replacing in a top level statement
                            assert parent.expr == old
                            parent.expr = new
                    return canReplace
            else:
                expr = args
                if hasSideEffects(expr):
                    return False
        return False

    def visitReplace(scope):
        newstatements = []
        for item in reversed(scope.statements):
            for sub in item.getScopes():
                visitReplace(sub)

            if isinstance(item.expr, ast.Assignment) and item.expr.params[0] in replacevars:
                expr_roots = []
                for item2 in newstatements:
                    # Don't inline into a while condition as it may be evaluated more than once
                    if not isinstance(item2, ast.WhileStatement):
                        expr_roots.append((item2, item2.expr))
                    if item2.getScopes():
                        break
                success = doReplacement(item, expr_roots)
                if success:
                    continue
            newstatements.insert(0, item)
        scope.statements = newstatements
    visitReplace(root)

def _createDeclarations(root, predeclared):
    _setScopeParents(root)
    info = findVarDeclInfo(root, predeclared)
    localdefs = collections.defaultdict(list)
    newvars = [var for var in info if isinstance(var, ast.Local) and info[var].declScope is None]
    remaining = set(newvars)

    def mdVisitVarUse(var):
        decl = ast.VariableDeclarator(ast.TypeName(var.dtype), var)
        # The compiler treats statements as if they can throw any exception at any time, so
        # it may think variables are not definitely assigned even when they really are.
        # Therefore, we give an unused initial value to every variable declaration
        # TODO - find a better way to handle this
        right = ast.dummyLiteral(var.dtype)
        localdefs[info[var].scope].append(ast.LocalDeclarationStatement(decl, right))
        remaining.remove(var)

    def mdVisitScope(scope):
        if isinstance(scope, ast.StatementBlock):
            for i,stmt in enumerate(scope.statements):
                if isinstance(stmt, ast.ExpressionStatement):
                    if isinstance(stmt.expr, ast.Assignment):
                        var, right = stmt.expr.params
                        if var in remaining and scope == info[var].scope:
                            decl = ast.VariableDeclarator(ast.TypeName(var.dtype), var)
                            new = ast.LocalDeclarationStatement(decl, right)
                            scope.statements[i] = new
                            remaining.remove(var)
                if stmt.expr is not None:
                    top = stmt.expr
                    for expr in top.postFlatIter():
                        if expr in remaining:
                            mdVisitVarUse(expr)
                for sub in stmt.getScopes():
                    mdVisitScope(sub)

    mdVisitScope(root)
    # print remaining
    assert not remaining
    assert None not in localdefs
    for scope, ldefs in localdefs.items():
        scope.statements = ldefs + scope.statements

def _createTernaries(scope, item):
    if isinstance(item, ast.IfStatement) and len(item.getScopes()) == 2:
        block1, block2 = item.getScopes()

        if (len(block1.statements) == len(block2.statements) == 1) and block1.jumpKey == block2.jumpKey:
            s1, s2 = block1.statements[0], block2.statements[0]
            e1, e2 = s1.expr, s2.expr

            if isinstance(s1, ast.ReturnStatement) and isinstance(s2, ast.ReturnStatement):
                expr = None if e1 is None else ast.Ternary(item.expr, e1, e2)
                item = ast.ReturnStatement(expr, s1.tt)
            elif isinstance(s1, ast.ExpressionStatement) and isinstance(s2, ast.ExpressionStatement):
                if isinstance(e1, ast.Assignment) and isinstance(e2, ast.Assignment):
                    # if e1.params[0] == e2.params[0] and max(e1.params[1].complexity(), e2.params[1].complexity()) <= 1:
                    if e1.params[0] == e2.params[0]:
                        expr = ast.Ternary(item.expr, e1.params[1], e2.params[1])
                        temp = ast.ExpressionStatement(ast.Assignment(e1.params[0], expr))

                        if not block1.doesFallthrough():
                            assert not block2.doesFallthrough()
                            item = ast.StatementBlock(item.func, item.continueKey, item.breakKey, [temp], block1.jumpKey)
                        else:
                            item = temp
    if item.expr is not None:
        item.expr = _simplifyExpressions(item.expr)
    return [item]

def _fixExprStatements(scope, item, namegen):
    if isinstance(item, ast.ExpressionStatement):
        right = item.expr
        if not isinstance(right, (ast.Assignment, ast.ClassInstanceCreation, ast.MethodInvocation)) and right.dtype is not None:
            left = ast.Local(right.dtype, lambda expr:namegen.getPrefix('dummy'))
            decl = ast.VariableDeclarator(ast.TypeName(left.dtype), left)
            item = ast.LocalDeclarationStatement(decl, right)
    return [item]

def _fixLiterals(scope, item):
    item.fixLiterals()

def _addCastsAndParens(scope, item, env):
    item.addCastsAndParens(env)

def _fallsThrough(scope, usedBreakTargets):
    # Check if control reaches end of scope and there is no break statement
    # We don't have to check keys since breaks should have already been generated for child scopes
    # for main scope there won't be one yet, but we don't care since we're just looking for
    # whether end of scope is reached on the main scope
    if not scope.statements:
        return True
    last = scope.statements[-1]
    if isinstance(last, (ast.JumpStatement, ast.ReturnStatement, ast.ThrowStatement)):
        return False
    elif not last.getScopes():
        return True
    # Scope ends with a complex statement. Determine whether it can fallthrough
    if last in usedBreakTargets:
        return True
    # Less strict than Java reachability rules, but we aren't bothering to follow them exactly
    if isinstance(last, ast.WhileStatement):
        return last.expr != ast.Literal.TRUE
    elif isinstance(last, ast.SwitchStatement):
        return not last.hasDefault() or _fallsThrough(last.getScopes()[-1], usedBreakTargets)
    else:
        if isinstance(last, ast.IfStatement) and len(last.getScopes()) < 2:
            return True
        return any(_fallsThrough(sub, usedBreakTargets) for sub in last.getScopes())

def _chooseJump(choices, breakPair, continuePair):
    assert None not in choices
    if breakPair in choices:
        return breakPair
    if continuePair in choices:
        return continuePair
    # Try to find an already labeled target
    for b, t in choices:
        if b.label is not None:
            return b, t
    return choices[0]

def _generateJumps(scope, usedBreakTargets, targets=collections.defaultdict(tuple), breakPair=None, continuePair=None, fallthroughs=NONE_SET, dryRun=False):
    assert None in fallthroughs
    newfallthroughs = fallthroughs
    newcontinuePair = continuePair
    newbreakPair = breakPair

    if scope.jumpKey not in fallthroughs:
        newfallthroughs = frozenset([None, scope.jumpKey])

    for item in reversed(scope.statements):
        if not item.getScopes():
            newfallthroughs = NONE_SET
            continue

        if isinstance(item, ast.WhileStatement):
            newfallthroughs = frozenset([None, item.continueKey])
        else:
            newfallthroughs |= frozenset([item.breakKey])

        newtargets = targets.copy()
        if isinstance(item, ast.WhileStatement):
            newtargets[item.continueKey] += ((item, True),)
            newcontinuePair = item, True
        if isinstance(item, (ast.WhileStatement, ast.SwitchStatement)):
            newbreakPair = item, False
        newtargets[item.breakKey] += ((item, False),)

        for subscope in reversed(item.getScopes()):
            _generateJumps(subscope, usedBreakTargets, newtargets, newbreakPair, newcontinuePair, newfallthroughs, dryRun=dryRun)
            if isinstance(item, ast.SwitchStatement):
                newfallthroughs = frozenset([None, subscope.continueKey])
        newfallthroughs = frozenset([None, item.continueKey])

    for item in scope.statements:
        if isinstance(item, ast.StatementBlock) and item.statements:
            if isinstance(item.statements[-1], ast.JumpStatement):
                assert item is scope.statements[-1] or item in usedBreakTargets

    # Now that we've visited children, decide if we need a jump for this scope, and if so, generate it
    if scope.jumpKey not in fallthroughs:
        # Figure out if this jump is actually reachable
        if _fallsThrough(scope, usedBreakTargets):
            target, isContinue = pair = _chooseJump(targets[scope.jumpKey], breakPair, continuePair)
            if not isContinue:
                usedBreakTargets.add(target)
            if pair == breakPair or pair == continuePair:
                target = None
            # Now actually add the jump statement
            if not dryRun:
                scope.statements.append(ast.JumpStatement(target, isContinue))

def _pruneVoidReturn(scope):
    if scope.statements:
        last = scope.statements[-1]
        if isinstance(last, ast.ReturnStatement) and last.expr is None:
            scope.statements.pop()

def _pruneCtorSuper(scope):
    if scope.statements:
        stmt = scope.statements[0]
        if isinstance(stmt, ast.ExpressionStatement):
            expr = stmt.expr
            if isinstance(expr, ast.MethodInvocation) and len(expr.params) == 0 and expr.name == 'super':
                scope.statements = scope.statements[1:]

def generateAST(method, graph, forbidden_identifiers):
    env = method.class_.env
    namegen = NameGen(forbidden_identifiers)
    class_ = method.class_
    inputTypes = parseMethodDescriptor(method.descriptor, unsynthesize=False)[0]
    tts = objtypes.verifierToSynthetic_seq(inputTypes)

    if graph is not None:
        graph.splitDualInedges()
        graph.fixLoops()
        entryNode, nodes = graphproxy.createGraphProxy(graph)
        if not method.static:
            entryNode.invars[0].name = 'this'

        setree = structuring.structure(entryNode, nodes, (method.name == '<clinit>'))
        ast_root, varinfo = astgen.createAST(method, graph, setree, namegen)

        argsources = [varinfo.var(entryNode, var) for var in entryNode.invars]
        disp_args = argsources if method.static else argsources[1:]
        for expr, tt in zip(disp_args, tts):
            expr.dtype = tt

        decls = [ast.VariableDeclarator(ast.TypeName(expr.dtype), expr) for expr in disp_args]
        ################################################################################################
        ast_root.bases = (ast_root,) # needed for our setScopeParents later

        assert _generateJumps(ast_root, set(), dryRun=True) is None
        _preorder(ast_root, _fixObjectCreations)
        boolize.boolizeVars(ast_root, argsources)
        boolize.interfaceVars(env, ast_root, argsources)
        _simplifyBlocks(ast_root)
        assert _generateJumps(ast_root, set(), dryRun=True) is None

        _mergeVariables(ast_root, argsources, method.static)

        _preorder(ast_root, _createTernaries)
        _inlineVariables(ast_root)
        _simplifyBlocks(ast_root)
        _preorder(ast_root, _createTernaries)
        _inlineVariables(ast_root)
        _simplifyBlocks(ast_root)

        _createDeclarations(ast_root, argsources)
        _preorder(ast_root, partial(_fixExprStatements, namegen=namegen))
        _preorder(ast_root, _fixLiterals)
        _preorder(ast_root, partial(_addCastsAndParens, env=env))
        _generateJumps(ast_root, set())
        _pruneVoidReturn(ast_root)
        _pruneCtorSuper(ast_root)
    else: # abstract or native method
        ast_root = None
        argsources = [ast.Local(tt, lambda expr:namegen.getPrefix('arg')) for tt in tts]
        decls = [ast.VariableDeclarator(ast.TypeName(expr.dtype), expr) for expr in argsources]

    flags = method.flags - set(['BRIDGE','SYNTHETIC','VARARGS'])
    if method.name == '<init>': # More arbtirary restrictions. Yay!
        flags = flags - set(['ABSTRACT','STATIC','FINAL','NATIVE','STRICTFP','SYNCHRONIZED'])

    flagstr = ' '.join(map(str.lower, sorted(flags)))
    inputTypes, returnTypes = parseMethodDescriptor(method.descriptor, unsynthesize=False)
    ret_tt = objtypes.verifierToSynthetic(returnTypes[0]) if returnTypes else objtypes.VoidTT
    return ast2.MethodDef(class_, flagstr, method.name, method.descriptor, ast.TypeName(ret_tt), decls, ast_root)



import os.path
import zipfile

from .classfile import ClassFile
from .classfileformat.reader import Reader
from .error import ClassLoaderError

class Environment(object):
    def __init__(self):
        self.classes = {}
        self.path = []
        self._open = {}

    def addToPath(self, path):
        self.path.append(path)

    def _getSuper(self, name):
        return self.getClass(name).supername

    def getClass(self, name, partial=False):
        try:
            result = self.classes[name]
        except KeyError:
            result = self._loadClass(name)
        if not partial:
            result.loadElements()
        return result

    def isSubclass(self, name1, name2):
        if name2 == 'java/lang/Object':
            return True

        while name1 != 'java/lang/Object':
            if name1 == name2:
                return True
            name1 = self._getSuper(name1)
        return False

    def commonSuperclass(self, name1, name2):
        a, b = name1, name2
        supers = {a}
        while a != b and a != 'java/lang/Object':
            a = self._getSuper(a)
            supers.add(a)

        while b not in supers:
            b = self._getSuper(b)
        return b

    def isInterface(self, name, forceCheck=False):
        try:
            class_ = self.getClass(name, partial=True)
            return 'INTERFACE' in class_.flags
        except ClassLoaderError as e:
            if forceCheck:
                raise e
            # If class is not found, assume worst case, that it is a interface
            return True

    def isFinal(self, name):
        try:
            class_ = self.getClass(name, partial=True)
            return 'FINAL' in class_.flags
        except ClassLoaderError as e:
            return False

    def _searchForFile(self, name):
        name += '.class'
        for place in self.path:
            try:
                archive = self._open[place]
            except KeyError: # plain folder
                try:
                    path = os.path.join(place, name)
                    with open(path, 'rb') as file_:
                        return file_.read()
                except IOError:
                    print('failed to open %s' % path.encode('utf8'))
            else: # zip archive
                try:
                    return archive.read(name)
                except KeyError:
                    pass

    def _loadClass(self, name):
        print("Loading %s" % name[:70])
        data = self._searchForFile(name)

        if data is None:
            raise ClassLoaderError('ClassNotFoundException', name)

        stream = Reader(data=data)
        new = ClassFile(stream)
        new.env = self
        self.classes[new.name] = new
        return new

    # Context Manager methods to manager our zipfiles
    def __enter__(self):
        assert not self._open
        for place in self.path:
            if place.endswith('.jar') or place.endswith('.zip'):
                self._open[place] = zipfile.ZipFile(place, 'r').__enter__()
        return self

    def __exit__(self, type_, value, traceback):
        for place in reversed(self.path):
            if place in self._open:
                self._open[place].__exit__(type_, value, traceback)
                del self._open[place]



from .. import excepttypes, objtypes, ssa_types
from ..constraints import IntConstraint, ObjectConstraint, join, maybeThrow, return_, throw

from .base import BaseOp

class CheckCast(BaseOp):
    def __init__(self, parent, target, args):
        super(CheckCast, self).__init__(parent, args, makeException=True)
        self.env = parent.env
        self.target_tt = target
        # Temporary hack
        target = objtypes.removeInterface(self.env, target)

        if objtypes.isBaseTClass(target):
            self.outCasted = ObjectConstraint.fromTops(parent.env, [target], [])
        else:
            # Primative array types need to be in exact, not supers
            self.outCasted = ObjectConstraint.fromTops(parent.env, [], [target])
        self.outExceptionCons = ObjectConstraint.fromTops(parent.env, [], (excepttypes.ClassCast,), nonnull=True)

    def propagateConstraints(self, x):
        intersect = join(x, self.outCasted)
        if intersect is None:
            return throw(self.outExceptionCons)
        elif intersect != x:
            assert not x.isConstNull()
            return maybeThrow(self.outExceptionCons)
        else:
            return return_(None)

class InstanceOf(BaseOp):
    def __init__(self, parent, target, args):
        super(InstanceOf, self).__init__(parent, args)
        self.env = parent.env
        self.target_tt = target
        self.rval = parent.makeVariable(ssa_types.SSA_INT, origin=self)

    def propagateConstraints(self, x):
        rvalcons = IntConstraint.range(32, 0, 1)
        return return_(rvalcons)
